"""user.user_id -> user.id

Revision ID: 74ab69a8dcb6
Revises: ffe998929da4
Create Date: 2018-03-16 12:35:12.742624

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '74ab69a8dcb6'
down_revision = 'ffe998929da4'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.alter_column('tUser', 'user_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)


    else:
        try:
            op.drop_table('tUser_new')
        except:
            log.info('Table tUser_new does not exist')

        try:
            op.create_table(
                'tUser_new',
                sa.Column('id', sa.Integer, primary_key=True, nullable=False),
                sa.Column('username', sa.Text(60),  nullable=False, unique=True),
                sa.Column('password', sa.Text(1000),  nullable=False),
                sa.Column('display_name', sa.Text(60),  nullable=False, server_default=text(u"''")),
                sa.Column('last_login', sa.TIMESTAMP()),
                sa.Column('last_edit', sa.TIMESTAMP()),
                sa.Column('cr_date', sa.TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tUser_new (id, username, password, display_name, last_login, last_edit, cr_date) select user_id, username, password, display_name, last_login, last_edit, cr_date from tUser")

            op.rename_table('tUser', 'tuser_old')
            op.rename_table('tUser_new', 'tUser')
            op.drop_table('tUser_old')
        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.alter_column('tUser', 'user_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)
    else:
        try:
            op.drop_table('tUser_new')
        except:
            log.info('Table tUser_new does not exist')

        try:
            op.create_table(
                'tUser_new',
                sa.Column('user_id', sa.Integer, primary_key=True, nullable=False),
                sa.Column('username', sa.Text(60),  nullable=False, unique=True),
                sa.Column('password', sa.Text(1000),  nullable=False),
                sa.Column('display_name', sa.Text(60),  nullable=False, server_default=text(u"''")),
                sa.Column('last_login', sa.TIMESTAMP()),
                sa.Column('last_edit', sa.TIMESTAMP()),
                sa.Column('cr_date', sa.TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tUser_new (user_id, username, password, display_name, last_login, last_edit, cr_date) select id, username, password, display_name, last_login, last_edit, cr_date from tUser")

            op.rename_table('tUser', 'tuser_old')
            op.rename_table('tUser_new', 'tUser')
            op.drop_table('tUser_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

#***************************************************
#Ownership & Permissions
#***************************************************
__all__ = ['Perm', 'Role', 'RolePerm', 'RoleUser', 'User']

class Perm(Base, Inspect):
    """
    """

    __tablename__='tPerm'

    id = Column(Integer(), primary_key=True, nullable=False)
    code = Column(String(60),  nullable=False)
    name = Column(String(200),  nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    _parents  = ['tRole', 'tPerm']
    _children = []

    def __repr__(self):
        return "{0} ({1})".format(self.name, self.code)

class Role(Base, Inspect):
    """
    """

    __tablename__='tRole'

    id = Column(Integer(), primary_key=True, nullable=False)
    code = Column(String(60),  nullable=False)
    name = Column(String(200),  nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    _parents  = []
    _children = ['tRolePerm', 'tRoleUser']

    @property
    def permissions(self):
        return set([rp.perm for rp in self.roleperms])

    def __repr__(self):
        return "{0} ({1})".format(self.name, self.code)


class RolePerm(Base, Inspect):
    """
    """

    __tablename__ = 'tRolePerm'

    perm_id = Column(Integer(), ForeignKey('tPerm.id'), primary_key=True, nullable=False)
    role_id = Column(Integer(), ForeignKey('tRole.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    perm = relationship('Perm', backref=backref('roleperms', uselist=True, lazy='joined'), lazy='joined')
    role = relationship('Role', backref=backref('roleperms', uselist=True, lazy='joined'), lazy='joined')

    _parents = ['tRole', 'tPerm']
    _children = []

    def __repr__(self):
        return "{0}".format(self.perm)

class RoleUser(Base, Inspect):
    """
    """

    __tablename__='tRoleUser'

    user_id = Column(Integer(), ForeignKey('tUser.id'), primary_key=True, nullable=False)
    role_id = Column(Integer(), ForeignKey('tRole.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    role = relationship('Role', backref=backref('roleusers', uselist=True))
    user = relationship('User', backref=backref('roleusers', uselist=True))

    _parents  = ['tRole', 'tUser']
    _children = []

    def __repr__(self):
        return "{0}".format(self.role.name)


class User(Base, Inspect):
    """
    """

    __tablename__='tUser'

    id = Column(Integer(), primary_key=True, nullable=False)
    username = Column(String(60),  nullable=False, unique=True)
    password = Column(LargeBinary(),  nullable=False)
    display_name = Column(String(200),  nullable=False, server_default=text(u"''"))
    last_login = Column(TIMESTAMP())
    last_edit = Column(TIMESTAMP())
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    failed_logins = Column(SMALLINT, nullable=True, default=0)

    _parents  = []
    _children = ['tRoleUser']

    def validate_password(self, password):
        if bcrypt.hashpw(password.encode('utf-8'), self.password.encode('utf-8')) == self.password.encode('utf-8'):
            return True
        return False

    @property
    def permissions(self):
        """Return a set with all permissions granted to the user."""
        perms = set()
        for r in self.roles:
            perms = perms | set(r.permissions)
        return perms

    @property
    def roles(self):
        """Return a set with all roles granted to the user."""
        roles = []
        for ur in self.roleusers:
            roles.append(ur.role)
        return set(roles)

    def is_admin(self):
        """
            Check that the user has a role with the code 'admin'
        """
        for ur in self.roleusers:
            if ur.role.code == 'admin':
                return True

        return False

    def __repr__(self):
        return "{0}".format(self.username)


#!/usr/bin/python3
"""
    This script print a report pointing the missing test files, and for the existing files the missing test methods
"""
from os import walk
import re

source_folder="./hydra_base/lib"
test_folder="./tests"

def check_correct_file(fileName):
    # Extracts files that are ".py" and not starting with _
    match = re.match( r'[^_].*\.py$', fileName, re.I)
    if match is not None:
        return True
    else:
        return False


def extract_file_names_from_dir(folder):
    # Extracts files names from a directory
    for (dirpath, dirnames, filenames) in walk(folder):
        break
    files = list(filter(check_correct_file, filenames ))
    files.sort()
    return files


source_files = extract_file_names_from_dir(source_folder)
test_files = extract_file_names_from_dir(test_folder)

source_methods = dict()
test_methods = {}

def extract_methods_from_file(folder_name, file_name):
    # Returns all the methods int he file that are not private to the file
    file_methods = []
    with open("{}/{}".format(folder_name,file_name)) as f:
        content = f.readlines()
        content = [x.strip() for x in content]
        for line in content:
            match = re.match( r'^def (OLD_)?([^_].*)\(', line)
            if match is not None:
                file_methods.append(match.group(2))
            else:
                pass

    file_methods.sort()
    return file_methods

# Getting all methods for all source files
for source_filename in source_files:
    source_methods[source_filename] = extract_methods_from_file(source_folder, source_filename)

# Getting all methods for all test files
for test_filename in test_files:
    test_methods[test_filename] = extract_methods_from_file(test_folder, test_filename)

# Comparing the existence of test methods for each source file
print
for source_file in source_methods:
    test_file_name = "test_{}".format(source_file)
    if not test_file_name in test_methods:
        # The test filename does not exists
        print ("Filename:{}, Missing: 'ALL'".format(test_file_name))
    else:
        test_file_methods = test_methods[test_file_name]
        for source_method in source_methods[source_file]:
            test_method = "test_{}".format(source_method)
            if not test_method in test_file_methods:
                print ("File:{}, Missing: {}".format(test_file_name, test_method))
                #print("The test called '{}' in the test filename {} does not exists!".format(test_method, test_file_name))



#s!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from lxml import etree
from hydra_base import config
from hydra_base.lib.objects import JSONObject
import hydra_base as hb
import logging
from hydra_base.exceptions import HydraError
import datetime
import json
import os
import pytest
log = logging.getLogger(__name__)


@pytest.fixture()
def template():
    return os.path.join(os.path.dirname(__file__), 'template.xml')


@pytest.fixture()
def template_json_object(client, template):

    with open(template) as fh:
        file_contents = fh.read()

    return JSONObject(client.import_template_xml(file_contents))


@pytest.fixture()
def mock_template(client):
    link_attr_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
    link_attr_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')
    node_attr_1 = client.testutils.create_attribute("node_attr_1", dimension='Volume')
    node_attr_2 = client.testutils.create_attribute("node_attr_2", dimension='Speed')
    net_attr_1 = client.testutils.create_attribute("net_attr_2", dimension='Speed')

    template = JSONObject()
    template.name = 'Test template @ %s' % datetime.datetime.now()

    layout = {}
    layout['groups'] = '<groups>...</groups>'

    template.layout = layout

    template.templatetypes = []
    # **********************
    # type 1           #
    # **********************
    type1 = JSONObject()
    type1.name = "Node type"
    type1.alias = "Node type alias"
    type1.resource_type = 'NODE'

    type1.typeattrs = []

    tattr_1 = JSONObject()
    tattr_1.attr_id = node_attr_1.id
    tattr_1.description = "Type attribute 1 description"
    tattr_1.properties = {'test_property': "test property add type"}
    tattr_1.data_restriction = {'LESSTHAN': 10, 'NUMPLACES': 1}
    type1.typeattrs.append(tattr_1)

    tattr_2 = JSONObject()
    tattr_2.attr_id = node_attr_2.id
    tattr_2.description = "Type attribute 2 description"
    tattr_2.data_restriction = {'INCREASING': None}
    type1.typeattrs.append(tattr_2)

    template.templatetypes.append(type1)
    # **********************
    # type 2           #
    # **********************
    type2 = JSONObject()
    type2.name = "Link type"
    type2.alias = "Link type alias"
    type2.resource_type = 'LINK'

    type2.typeattrs = []

    tattr_1 = JSONObject()
    tattr_1.attr_id = link_attr_1.id
    type2.typeattrs.append(tattr_1)

    tattr_2 = JSONObject()
    tattr_2.attr_id = link_attr_2.id
    type2.typeattrs.append(tattr_2)

    template.templatetypes.append(type2)

    # **********************
    # type 3           #
    # **********************
    type3 = JSONObject()
    type3.name = "Network Type"
    type3.alias = "Network Type alias"
    type3.resource_type = 'NETWORK'

    type3.typeattrs = []

    tattr_3 = JSONObject()
    tattr_3.attr_id = net_attr_1.id
    tattr_3.data_restriction = {}

    type3.typeattrs.append(tattr_3)

    template.templatetypes.append(type3)

    new_template_i = client.add_template(template)
    # TODO: HACK to load the attr
    for tt in new_template_i.templatetypes:
        for ta in tt.typeattrs:
            ta.attr

    new_template_j = JSONObject(new_template_i)
    return new_template_j

# mock_template_copy is used for some of compatiblity tests but is identical to the
# mock_template fixture
mock_template_copy = mock_template


class TestTemplates:
    """
        Test for templates
    """
    def test_add_xml(self, template_json_object):
        new_tmpl = template_json_object

        assert new_tmpl is not None, "Adding template from XML was not successful!"

        assert len(new_tmpl.templatetypes) == 2

        for tt in new_tmpl.templatetypes:
            if tt.name == 'Reservoir':
                for ta in tt.typeattrs:
                    assert ta.data_type == 'scalar'

                assert tt.typeattrs[-1].properties is not None
                assert eval(tt.typeattrs[-1].properties)['template_property'] == "Test property from template"

    def test_get_xml(self, client, template_json_object):
        xml_tmpl = template_json_object

        db_template = client.get_template_as_xml(xml_tmpl.id)


        assert db_template is not None

        template_xsd_path = config.get('templates', 'template_xsd_path')
        xmlschema_doc = etree.parse(template_xsd_path)

        xmlschema = etree.XMLSchema(xmlschema_doc)

        xml_tree = etree.fromstring(db_template)

        xmlschema.assertValid(xml_tree)

    def test_get_dict(self, client, template_json_object):

        # Upload the xml file initally to avoid having to manage 2 template files
        xml_tmpl = template_json_object

        template_dict = client.get_template_as_dict(xml_tmpl.id)

        # Error that there's already a template with this name.
        with pytest.raises(HydraError):
            client.import_template_dict(template_dict, allow_update=False)

        typename = template_dict['template']['templatetypes'][0]['name']

        template_dict['template']['templatetypes'][0].name = typename + "_updated"

        # Finds a template with this name and updates it to mirror this dict.
        # This includes deleting types if they're not in this dict.
        # Changing the name of a type has this effect, as a new template does not have
        # any reference to existing types in Hydra.

        updated_template = JSONObject(client.import_template_dict(template_dict))

        type_names = []

        for templatetype in updated_template.templatetypes:
            type_names.append(templatetype.name)

        assert len(type_names) == 2
        assert typename + "_updated" in type_names and typename not in type_names

        # Now put it back to the original name so other tests will work
        log.info("Reverting the type's name")
        template_dict['template']['templatetypes'][0].name = typename
        updated_template = JSONObject(client.import_template_dict(template_dict))

        #just double-check that the JSON import works also
        updated_template = JSONObject(client.import_template_json(json.dumps(template_dict)))

        type_names = []
        for templatetype in updated_template.templatetypes:
            type_names.append(templatetype.name)

        assert len(type_names) == 2
        assert typename in type_names and typename + "_updated" not in type_names

        log.info("Checking to ensure Template has been updated correctly...")
        # one final check to ensure that the type has been deleted
        check_template_i = client.get_template(updated_template.id)

        assert len(check_template_i.templatetypes) == 2


    """
        TEMPLATES Functions
    """

    def test_add_template(self, client, mock_template):

        link_attr_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
        link_attr_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')
        node_attr_1 = client.testutils.create_attribute("node_attr_1", dimension='Volume')
        node_attr_2 = client.testutils.create_attribute("node_attr_2", dimension='Speed')

        new_template_j = mock_template

        assert new_template_j.name == mock_template.name, "Names are not the same!"
        assert json.dumps(new_template_j.layout) == json.dumps(mock_template.layout), "Layouts are not the same!"
        assert new_template_j.id is not None, "New Template has no ID!"
        assert new_template_j.id > 0, "New Template has incorrect ID!"

        assert len(new_template_j.templatetypes) == 3, "Resource types did not add correctly"
        for t in new_template_j.templatetypes[0].typeattrs:
            assert t.attr_id in (node_attr_1.id, node_attr_2.id);
            "Node types were not added correctly!"

        for t in new_template_j.templatetypes[1].typeattrs:
            assert t.attr_id in (link_attr_1.id, link_attr_2.id);
            "Node types were not added correctly!"

    def test_update_template(self, client):

        #Only applicable for tests.
        client.template.ATTR_CACHE = {}

        # Defining attributes
        attribute_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
        attribute_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')
        attribute_3 = client.testutils.create_attribute("node_attr_1", dimension='Volume')

        # Defining template
        template = JSONObject()

        template.name = 'Test Template @ %s'%datetime.datetime.now()

        template.templatetypes = []

        template_type_1 = JSONObject({
            "name" : "Node type 2",
            "alias" : "Node type 2 alias",
            "resource_type" : 'NODE',
            "typeattrs" : []
        })

        template_type_2 = JSONObject({
            "name" : "Link type 2",
            "alias" : "Link type 2 alias",
            "resource_type" : 'LINK',
            "typeattrs" : []
        })

        # Definind a typeattr data structure
        type_attribute_1 = JSONObject({
            "attr_id" : attribute_1.id,
            "unit_id" : client.get_unit_by_abbreviation('bar').id,
            "description" : "typeattr description 1",
            "properties" : {"test_property": "property value"}
        })
        template_type_1.typeattrs.append(type_attribute_1)

        type_attribute_2 = JSONObject({
            "attr_id" : attribute_2.id,
            "unit_id" : client.get_unit_by_abbreviation('mph').id,
            "description" : "typeattr description 2"
        })
        template_type_2.typeattrs.append(type_attribute_2)

        template.templatetypes.append(template_type_1)
        template.templatetypes.append(template_type_2)

        new_template_i = client.add_template(template)
        new_template_j = JSONObject(new_template_i)

        assert new_template_j.name == template.name, "Names are not the same!"
        assert new_template_j.id is not None, "New Template has no ID!"
        assert new_template_j.id > 0, "New Template has incorrect ID!"

        assert len(new_template_j.templatetypes) == 2, "Resource types did not add correctly"
        assert len(new_template_j.templatetypes[0].typeattrs) == 1, "Resource type attrs did not add correctly"
        assert new_template_j.templatetypes[0].typeattrs[0].unit_id == client.get_unit_by_abbreviation('bar').id

        #update the name of one of the types
        new_template_j.templatetypes[0].name = "Test type 3"
        updated_type_id = new_template_j.templatetypes[0].id

        #add a new type
        new_type = JSONObject({
            "name" : "New Node type",
            "alias" : "New Node type alias",
            "resource_type" : 'NODE',
            "typeattrs" : []
        })
        new_template_j.templatetypes.append(new_type)

        #add an template attr to one of the types
        type_attribute_3 = JSONObject({
            "attr_id" : attribute_3.id,
            "description" : "updated typeattr description 1",
            "properties" : {"test_property_of_added_type": "property value"}
        })
        new_template_j.templatetypes[0].typeattrs.append(type_attribute_3)
        updated_template_j = client.update_template(new_template_j)
        updated_template_retrieved_j = client.get_template(new_template_j.id)

        assert updated_template_retrieved_j.name == template.name, "Names are not the same!"

        updated_type = None
        for tmpltype in new_template_j.templatetypes:
            if tmpltype.id == updated_type_id:
                updated_type = tmpltype
                break

        assert updated_type.name == "Test type 3", "Resource types did not update correctly"

        assert len(updated_type.typeattrs) == 2, "Resource type template attr did not update correctly"

        """
            Test that when setting a unit on a type attr, it matches the dimension of its attr
            In this case, setting m^3(Volume) fails as the attr has a dimension of 'Pressure'
        """
        old_attribute_type_to_restore = updated_template_j.templatetypes[0].typeattrs[0].unit_id
        updated_template_j.templatetypes[0].typeattrs[0].unit_id = client.get_unit_by_abbreviation('m^3').id
        with pytest.raises(HydraError):
            client.update_template(updated_template_j)
        # Restoring the old value to not fail later for this reason
        updated_template_j.templatetypes[0].typeattrs[0].unit_id = old_attribute_type_to_restore


        """
            Testing the case in which the attribute has no dimension while the typeattr has UNIT_ID that is not none
        """
        attribute_dimension_none = client.testutils.create_attribute("node_attr_dimension_none", dimension=None)
        template_type_dimension_none = JSONObject({
            "name" : "Node type dimension_none",
            "alias" : "Node type dimension_none alias",
            "resource_type" : 'NODE',
            "typeattrs" : []
        })
        type_attribute_dimension_none = JSONObject({
            "attr_id" : attribute_dimension_none.id,
            "unit_id" : client.get_unit_by_abbreviation('bar').id,
            "description" : "type_attribute_dimension_none description 1",
            "properties" : {"test_property": "property value"}
        })
        template_type_dimension_none.typeattrs.append(type_attribute_dimension_none)
        updated_template_j.templatetypes.append(template_type_dimension_none)

        with pytest.raises(HydraError):
            client.update_template(updated_template_j) # It MUST fail because the unit_id is not consistent to dimension_id none

    def test_delete_template(self, client, network_with_data, mock_template):

        #Only applicable for tests. TODO: make this not rubbish.
        client.template.ATTR_CACHE = {}

        network = network_with_data
        new_template = mock_template

        retrieved_template_j = client.get_template(new_template.id)
        assert retrieved_template_j is not None

        client.apply_template_to_network(retrieved_template_j.id, network.id)

        updated_network = client.get_network(network.id)
        assert len(updated_network.types) == 2

        expected_net_type = None
        for t in new_template.templatetypes:
            if t.resource_type == 'NETWORK':
                expected_net_type = t.id

        network_types = [t.id for t in updated_network.types]

        assert expected_net_type in network_types

        #can't delete this because there's a resource type associated to one
        #of its types
        with pytest.raises(HydraError):
            client.delete_template(new_template.id)

        client.delete_template(new_template.id, delete_resourcetypes=True)

        with pytest.raises(HydraError):
            client.get_template(new_template.id)

        network_deleted_templatetypes = client.get_network(network.id)

        assert len(network_deleted_templatetypes.types) == 1

    def test_get_types_by_attr(self, client, network_with_data, mock_template):

        #Only applicable for tests. TODO: make this not rubbish.
        client.template.ATTR_CACHE = {}

        network = network_with_data
        new_template = mock_template

        retrieved_template_j = client.get_template(new_template.id)
        assert retrieved_template_j is not None

        client.apply_template_to_network(retrieved_template_j.id, network.id)

        updated_network = client.get_network(network.id)
        assert len(updated_network.types) == 2


        node_types = client.get_types_by_attr(updated_network.nodes[0], 'NODE')

        assert len(node_types) > 0

        #there could be lots of templates in the system, so the node should
        #match *at least* the number of node types, but may not be equal
        assert len(node_types) >= len(updated_network.nodes[0].types)



    """
        TEMPLATE TYPES Functions
    """

    def test_add_type(self, client, mock_template):

        template = mock_template

        templatetype = client.testutils.create_templatetype(template.id)

        new_type_i = client.add_templatetype(templatetype)
        new_type_j = JSONObject(new_type_i)

        assert new_type_j.name == templatetype.name, "Names are not the same!"
        assert new_type_j.alias == templatetype.alias, "Aliases are not the same!"
        assert new_type_j.layout == templatetype.layout, "Layouts are not the same!"
        assert new_type_j.id is not None, "New type has no ID!"
        assert new_type_j.id > 0, "New type has incorrect ID!"

        assert len(new_type_j.typeattrs) == 3, "Resource type attrs did not add correctly"

    def test_update_type(self, client, mock_template):

        template = mock_template

        attr_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
        attr_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')
        attr_3 = client.testutils.create_attribute("node_attr_1", dimension='Volume')

        templatetype = JSONObject()
        templatetype.name = "Test type name @ %s" % (datetime.datetime.now())
        templatetype.alias = templatetype.name + " alias"
        templatetype.template_id = template.id
        templatetype.resource_type = 'NODE'

        tattr_1 = JSONObject()
        tattr_1.attr_id = attr_1.id

        tattr_2 = JSONObject()
        tattr_2.attr_id = attr_2.id

        templatetype.typeattrs = [tattr_1, tattr_2]

        new_type_i = client.add_templatetype(templatetype)
        type_to_update_j = JSONObject(new_type_i)

        assert type_to_update_j.name == templatetype.name, "Names are not the same!"
        assert type_to_update_j.alias == templatetype.alias, "Aliases are not the same!"
        assert type_to_update_j.id is not templatetype, "New type has no ID!"
        assert type_to_update_j.id > 0, "New type has incorrect ID!"

        assert len(type_to_update_j.typeattrs) == 2, "Resource type attrs did not add correctly"
        type_to_update_j.name = "Updated type name @ %s"%(datetime.datetime.now())
        type_to_update_j.alias = templatetype.name + " alias"
        type_to_update_j.resource_type = 'NODE'

        tattr_3 = JSONObject()
        tattr_3.attr_id = attr_3.id
        tattr_3.description = "Descripton of added typeattr"
        tattr_3.properties = {"update_type_test_property": "property value"}
        type_to_update_j.typeattrs.append(tattr_3)

        type_to_update_j.typeattrs[0].description = "Updated typeattr description"

        updated_type_i = client.update_templatetype(type_to_update_j)
        updated_type_j = JSONObject(updated_type_i)

        assert type_to_update_j.name == updated_type_j.name, "Names are not the same!"
        assert type_to_update_j.alias == updated_type_j.alias, "Aliases are not the same!"
        assert type_to_update_j.id == updated_type_j.id, "type ids to not match!"
        assert type_to_update_j.id > 0, "New type has incorrect ID!"
        assert type_to_update_j.typeattrs[0].description == "Updated typeattr description"
        assert type_to_update_j.typeattrs[-1].properties['update_type_test_property'] == "property value"

        assert len(updated_type_j.typeattrs) == 3, "Template type attrs did not update correctly"

    def test_delete_type(self, client, mock_template):
        new_template = mock_template

        retrieved_template = client.get_template(new_template.id)
        assert retrieved_template is not None

        templatetype = new_template.templatetypes[0]
        client.delete_templatetype(templatetype.id)

        updated_template = JSONObject(client.get_template(new_template.id))

        for tmpltype in updated_template.templatetypes:
            assert tmpltype.id != templatetype.id

    def test_get_type(self, client, mock_template):
        new_type = mock_template.templatetypes[0]
        new_type = client.get_templatetype(new_type.id)
        assert new_type is not None, "Resource type attrs not retrived by ID!"

    def test_get_type_by_name(self, client, mock_template):
        up_to_date_template = client.get_template(mock_template.id)
        #Get this in case the mock template has been updated since creation
        type = up_to_date_template.templatetypes[0]
        new_type = client.get_templatetype_by_name(type.template_id, type.name)
        assert new_type is not None, "Resource type attrs not retrived by name!"

    def test_clone_templatetype(self, client, mock_template):
        """Test cloning a template type"""
        original_template = client.get_template(mock_template.id)
        original_type = original_template.templatetypes[0]

        # Clone the template type
        cloned_type = client.clone_templatetype(original_type.id)
        cloned_type_j = JSONObject(cloned_type)

        # Verify the clone was created successfully
        assert cloned_type_j is not None, "Cloned template type should not be None"
        assert cloned_type_j.id is not None, "Cloned template type should have an ID"
        assert cloned_type_j.id != original_type.id, "Cloned type should have different ID than original"

        # Verify that only the specified fields are different
        # Name should be modified to be unique (appends "(Clone)")
        assert cloned_type_j.name == original_type.name + " (Clone)", "Cloned type should have modified name"
        assert cloned_type_j.alias == original_type.alias, "Cloned type should have same alias as original"
        assert cloned_type_j.description == original_type.description, "Cloned type should have same description as original"
        assert cloned_type_j.resource_type == original_type.resource_type, "Cloned type should have same resource_type as original"
        assert cloned_type_j.template_id == original_type.template_id, "Cloned type should have same template_id as original"
        assert cloned_type_j.layout == original_type.layout, "Cloned type should have same layout as original"

        # Verify that typeattrs are cloned correctly
        assert len(cloned_type_j.typeattrs) == len(original_type.typeattrs), "Cloned type should have same number of typeattrs"

        # Check that typeattrs are properly cloned (different IDs but same content)
        for i, cloned_typeattr in enumerate(cloned_type_j.typeattrs):
            original_typeattr = original_type.typeattrs[i]

            assert cloned_typeattr.id != original_typeattr.id, "Cloned typeattr should have different ID"
            assert cloned_typeattr.type_id == cloned_type_j.id, "Cloned typeattr should reference cloned type"
            assert cloned_typeattr.attr_id == original_typeattr.attr_id, "Cloned typeattr should have same attr_id"
            assert cloned_typeattr.description == original_typeattr.description, "Cloned typeattr should have same description"
            assert cloned_typeattr.properties == original_typeattr.properties, "Cloned typeattr should have same properties"
            assert cloned_typeattr.data_restriction == original_typeattr.data_restriction, "Cloned typeattr should have same data_restriction"
            assert cloned_typeattr.unit_id == original_typeattr.unit_id, "Cloned typeattr should have same unit_id"

        # Test cloning multiple times to ensure unique names
        cloned_type_2 = client.clone_templatetype(original_type.id)
        cloned_type_2_j = JSONObject(cloned_type_2)

        assert cloned_type_2_j.name == original_type.name + " (Clone) 1", "Second clone should have incremented name"
        assert cloned_type_2_j.id != cloned_type_j.id, "Multiple clones should have different IDs"


    """
        Type Attributes Functions
    """


    def test_add_typeattr(self, client, mock_template):

        attr_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
        attr_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')
        attr_3 = client.testutils.create_attribute("node_attr_1", dimension='Volume')

        templatetype = JSONObject()
        templatetype.name = "Test type name @ %s"%(datetime.datetime.now())
        templatetype.alias = templatetype.name + " alias"
        templatetype.template_id = mock_template.id
        templatetype.resource_type = 'NODE'

        tattr_1 = JSONObject()
        tattr_1.attr_id = attr_1.id

        tattr_2 = JSONObject()
        tattr_2.attr_id = attr_2.id
        tattr_2.description = "Description of typeattr from test_add_typeattr"
        tattr_2.properties = {"test_property":"property value"}

        templatetype.typeattrs = [tattr_1, tattr_2]

        new_type = JSONObject(client.add_templatetype(templatetype))

        tattr_3 = JSONObject()
        tattr_3.attr_id = attr_3.id
        tattr_3.type_id = new_type.id
        tattr_3.description = "Description of additional typeattr from test_add_typeattr"
        tattr_3.properties = {"add_typeattr_test_property": "property value"}

        log.info("Adding Test Type attribute")

        client.add_typeattr(tattr_3)

        updated_type = JSONObject(client.get_templatetype(new_type.id))

        assert len(updated_type.typeattrs) == 3, "Type attr did not add correctly"

        assert eval(updated_type.typeattrs[-1].properties)['add_typeattr_test_property'] == "property value"

    def test_delete_typeattr(self, client, mock_template):

        template = mock_template

        attr_1 = client.testutils.create_attribute("link_attr_1", dimension='Pressure')
        attr_2 = client.testutils.create_attribute("link_attr_2", dimension='Speed')

        templatetype = JSONObject()
        templatetype.name = "Test type name @ %s"%(datetime.datetime.now())
        templatetype.alias = templatetype.name + " alias"
        templatetype.resource_type = 'NODE'
        templatetype.template_id = template.id

        tattr_1 = JSONObject()
        tattr_1.attr_id = attr_1.id

        tattr_2 = JSONObject()
        tattr_2.attr_id = attr_2.id

        templatetype.typeattrs = [tattr_1, tattr_2]

        new_type = JSONObject(client.add_templatetype(templatetype))

        assert len(new_type.typeattrs) == 2

        typeattr_to_delete = new_type.typeattrs[0]

        client.delete_typeattr(typeattr_to_delete.id)

        updated_type = JSONObject(client.get_templatetype(new_type.id))


        assert len(updated_type.typeattrs) == 1, "Resource type attr did not add correctly"

    def test_get_templates(self, client, mock_template):

        templates = [JSONObject(t) for t in client.get_templates()]
        for t in templates:
            for typ in t.templatetypes:
                assert typ.resource_type is not None
        assert len(templates) > 0, "Templates were not retrieved!"

    def test_get_template(self, client, mock_template):
        new_template = JSONObject(client.get_template(mock_template.id))

        assert new_template.name == mock_template.name, "Names are not the same! Retrieval by ID did not work!"

    def test_get_template_by_name_good(self, client, mock_template):
        new_template = JSONObject(client.get_template_by_name(mock_template.name))

        assert new_template.name == mock_template.name, "Names are not the same! Retrieval by name did not work!"

    def test_get_template_by_name_bad(self, client):

        with pytest.raises(HydraError):
            new_template = client.get_template_by_name("Not a template!")

    def test_set_template_status(self, client, network_with_data,  mock_template):

        all_templates = client.get_templates(load_all=False)
        assert len(all_templates) > 1

        tmpl_to_deactivate = all_templates[0].id

        client.deactivate_template(tmpl_to_deactivate)

        #this should return one fewer
        filtered_templates = client.get_templates(load_all=False)
        assert len(filtered_templates) == len(all_templates) - 1

        unfiltered_templates = client.get_templates(load_all=False, include_inactive=True)
        assert len(unfiltered_templates) == len(all_templates)

        client.activate_template(tmpl_to_deactivate)

        unfiltered_templates = client.get_templates(load_all=False)
        assert len(unfiltered_templates) == len(all_templates)

    """
        Resource Types Functions
    """

    def test_add_resource_type(self, client, mock_template):

        template = mock_template
        types = template.templatetypes
        type_name = types[0].name
        type_id   = types[0].id

        project = client.testutils.create_project()
        network = JSONObject()

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        network.nodes = []
        network.links = []

        for i in range(nnodes):
            node = JSONObject()
            node.id = i * -1
            node.name = 'Node ' + str(i)
            node.description = 'Test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            type_summary = JSONObject()
            type_summary.id = template.id
            type_summary.name = template.name
            type_summary.id = type_id
            type_summary.name = type_name

            node.types = [type_summary]

            network.nodes.append(node)

        for i in range(nlinks):
            link = JSONObject()
            link.id = i * -1
            link.name = 'Link ' + str(i)
            link.description = 'Test link ' + str(i)
            link.node_1_id = network.nodes[i].id
            link.node_2_id = network.nodes[i + 1].id

            network.links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'A network for SOAP unit tests.'

        net_summary = client.add_network(network)
        new_net = client.get_network(net_summary.id)

        for node in new_net.nodes:
            assert node.types is not None and node.types[0].name == "Node type"; "type was not added correctly!"

    def test_find_matching_resource_types(self, client, network_with_data):

        network = network_with_data

        node_to_check = network.nodes[0]
        matching_types_i = client.get_matching_resource_types('NODE', node_to_check.id)
        matching_types_j = [JSONObject(i) for i in matching_types_i]

        assert len(matching_types_j) > 0, "No types returned!"

        matching_type_ids = []
        for tmpltype in matching_types_j:
            matching_type_ids.append(tmpltype.id)

        assert node_to_check.types[0].id in matching_type_ids, "TemplateType ID not found in matching types!"

    def test_assign_type_to_resource(self, client, mock_template, network_with_data):
        network = network_with_data
        template = mock_template
        templatetype = template.templatetypes[0]

        node_to_assign = network.nodes[0]

        result = JSONObject(client.assign_type_to_resource(templatetype.id, 'NODE', node_to_assign.id))

        node = client.get_node(node_to_assign.id)

        assert len(node.types) == 2, 'Assigning type did not work properly.'

        assert str(result.id) in [str(x.type_id) for x in node.types]


    def test_assign_types_to_resources(self, client, mock_template, network_with_data):
        network = network_with_data
        template = mock_template
        templatetype = template.templatetypes[0]

        node_to_assign = network.nodes[0]

        resource_types = [JSONObject({
            'ref_key': 'NODE',
            'ref_id':node_to_assign.id,
            'type_id':templatetype.id
            })]
        results = client.assign_types_to_resources(resource_types)

        node = client.get_node(node_to_assign.id)

        assert len(node.types) == 2, 'Assigning type did not work properly.'

        assert str(results[0].id) in [str(x.type_id) for x in node.types]

    def test_remove_type_from_resource(self, client, mock_template, network_with_data):
        network = network_with_data
        template = mock_template
        templatetype = template.templatetypes[0]

        node_to_assign = network.nodes[0]

        result1_i = client.assign_type_to_resource(templatetype.id, 'NODE', node_to_assign.id)

        result1_j = JSONObject(result1_i)

        node_j = JSONObject(client.get_node(node_to_assign.id))

        assert node_j.types is not None, \
            'Assigning type did not work properly.'

        assert str(result1_j.id) in [str(x.type_id) for x in node_j.types]

        remove_result = client.remove_type_from_resource(templatetype.id, 'NODE', node_to_assign.id)
        assert remove_result == 'OK'

        updated_node_j = JSONObject(client.get_node(node_to_assign.id))

        assert updated_node_j.types is None or str(result1_j.id) not in [str(x.type_id) for x in updated_node_j.types]


    def test_create_template_from_network(self, client, network_with_data):
        network = network_with_data

        net_template = client.get_network_as_xml_template(network.id)


        assert net_template is not None

        template_xsd_path = config.get('templates', 'template_xsd_path')
        xmlschema_doc = etree.parse(template_xsd_path)

        xmlschema = etree.XMLSchema(xmlschema_doc)

        xml_tree = etree.fromstring(net_template)

        xmlschema.assertValid(xml_tree)

    def test_apply_template_to_network(self, client, mock_template, network_with_data):
        net_to_update = network_with_data
        template = mock_template

        # Test the links as it's easier
        empty_links = []
        for l in net_to_update.links:
            if l.types is None:
                empty_links.append(l.id)

        # Add the resource attributes to the links, so we can guarantee
        # that these links will match those in the template.
        for t in template.templatetypes:
            if t.resource_type == 'LINK':
                link_type = t
                break

        link_ra_1 = JSONObject(dict(
            attr_id=link_type.typeattrs[0].attr_id
        ))
        link_ra_2 = JSONObject(dict(
            attr_id=link_type.typeattrs[1].attr_id
        ))
        for link in net_to_update.links:
            if link.types is None:
                link.attributes.append(link_ra_1)
                link.attributes.append(link_ra_2)

        network = client.update_network(net_to_update)

        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == 'Default Node'

        client.apply_template_to_network(template.id, network.id)

        network = client.get_network(network.id)

        assert len(network.types) == 2
        assert 'Network Type' in [t.name for t in network.types]
        for l in network.links:
            if l.id in empty_links:
                assert l.types is not None
                assert len(l.types) == 1
                assert l.types[0].name == 'Link type'

        # THe assignment of the template hasn't affected the nodes
        # as they do not have the appropriate attributes.
        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == 'Default Node'

    def test_apply_template_to_network_twice(self, client, mock_template, network_with_data):
        net_to_update = network_with_data
        template = mock_template

        # Test the links as it's easier
        empty_links = []
        for l in net_to_update.links:
            if l.types is None:
                empty_links.append(l.id)

        # Add the resource attributes to the links, so we can guarantee
        # that these links will match those in the template.
        for t in template.templatetypes:
            if t.resource_type == 'LINK':
                link_type = t
                break

        link_ra_1 = JSONObject(dict(
            attr_id=link_type.typeattrs[0].attr_id
        ))
        link_ra_2 = JSONObject(dict(
            attr_id=link_type.typeattrs[1].attr_id
        ))
        for link in net_to_update.links:
            if link.types is None:
                link.attributes.append(link_ra_1)
                link.attributes.append(link_ra_2)

        network = client.update_network(net_to_update)

        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == 'Default Node'

        client.apply_template_to_network(template.id, network.id)

        client.apply_template_to_network(template.id, network.id)

        network = client.get_network(network.id)

        assert len(network.types) == 2

        assert 'Network Type' in [t.name for t in network.types]

        for l in network.links:
            if l.id in empty_links:
                assert l.types is not None
                assert len(n.types) == 1
                assert l.types[0].name == 'Link type'

        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == 'Default Node'


    def test_remove_template_from_network(self, client, network_with_data):
        network = network_with_data
        template_id = network.types[0].template_id

        # Test the links as it's easier
        empty_links = []
        for l in network.links:
            if l.types is None or len(l.types) == 0:
                empty_links.append(l.id)

        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == "Default Node"

        client.apply_template_to_network(template_id, network.id)

        client.remove_template_from_network(network.id, template_id, 'N')

        network_2 = client.get_network(network.id)

        assert len(network_2.types) == 0
        for l in network_2.links:
            if l.id in empty_links:
                assert len(l.types) == 0

        for n in network_2.nodes:
            assert len(n.types) == 0

    def test_remove_template_and_attributes_from_network(self, client, mock_template, network_with_data):
        network = network_with_data
        template = mock_template

        # Test the links as it's easier
        empty_links = []
        for l in network.links:
            if l.types is None:
                empty_links.append(l.id)

        for n in network.nodes:
            assert len(n.types) == 1
            assert n.types[0].name == 'Default Node'

        network_1 = client.get_network(network.id)
        assert len(network_1.types) == 1

        client.apply_template_to_network(template.id, network.id)

        network_3 = client.get_network(network.id)
        assert len(network_3.types) == 2

        client.remove_template_from_network(network.id, template.id, 'Y')

        network_2 = client.get_network(network.id)

        assert len(network_2.types) == 1

        link_attrs = []
        for tt in template.templatetypes:
            if tt.resource_type != 'LINK':
                continue
            for ta in tt.typeattrs:
                attr_id = ta.attr_id
                if attr_id not in link_attrs:
                    link_attrs.append(attr_id)
                link_attrs.append(ta.attr_id)
        for l in network_2.links:
            if l.id in empty_links:
                assert l.types is None
            if l.attributes is not None:
                for la in l.attributes:
                    assert la.attr_id not in link_attrs

        node_attrs = []
        for tt in template.templatetypes:
            if tt.resource_type != 'NODE':
                continue
            for ta in tt.typeattrs:
                attr_id = ta.attr_id
                if attr_id not in node_attrs:
                    node_attrs.append(attr_id)
                node_attrs.append(ta.attr_id)

        for n in network_2.nodes:
            assert len(n.types) == 1
            if n.attributes is not None:
                for node_attr in n.attributes:
                    assert node_attr.attr_id not in node_attrs

    def test_validate_attr(self, client, network_with_data):
        network = network_with_data

        scenario = network.scenarios[0]
        rs_ids = [rs.resource_attr_id for rs in scenario.resourcescenarios]
        template_id = network.nodes[0].types[0].template_id

        for n in network.nodes:
            node_type = client.get_templatetype(n.types[0].id)
            for ra in n.attributes:
                for type_attr in node_type.typeattrs:
                    if ra.attr_id == type_attr.attr_id and ra.id in rs_ids and type_attr.data_restriction is not None:
                #        logging.info("Validating RA %s in scenario %s", ra.id, scenario.id)
                        error = client.validate_attr(ra.id, scenario.id, template_id)
                        if error not in (None, {}):
                            assert error.ref_id == n.id


    def test_validate_attrs(self, client, network_with_data):
        network = network_with_data

        scenario = network.scenarios[0]

        ra_ids = []

        for rs in scenario.resourcescenarios:
            ra_ids.append(rs.resource_attr_id)

        template_id = network.types[0].template_id

        errors = client.validate_attrs(ra_ids, scenario.id, template_id)

        assert len(errors) > 0

    def test_validate_scenario(self, client, network_with_data):
        network = network_with_data

        scenario = network.scenarios[0]
        template_id = network.nodes[0].types[0].template_id

        errors = client.validate_scenario(scenario.id,template_id)

        assert len(errors) > 0

    def test_validate_network(self, client, network_with_data):
        network = network_with_data

        client.testutils.update_template(network.types[0].template_id)

        scenario = network.scenarios[0]
        networktype = network.nodes[0].types[0]
        # Validate the network without data: should pass as the network is built
        # based on the template in these unit tests
        errors1 = client.validate_network(network.id, networktype.template_id)
        # The network should have an error, saying that the template has net_attr_c,
        # but the network does not
        assert len(errors1) == 1
        assert errors1[0].find('net_attr_d') > 0

        # Validate the network with data. Should fail as one of the attributes (node_attr_3)
        # is specified as being a 'Cost' in the template but 'Speed' is the dimension
        # of the dataset used. In addition, node_attr_1 specified a unit of 'm^3'
        # whereas the timeseries in the data is in 'cm^3', so each will fail on unit
        # mismatch also.
        errors2 = client.validate_network(network.id, networktype.template_id,
                                                  scenario.id)

        assert len(errors2) > 0
        # every node should have an associated error, plus the network error from above
        assert len(errors2) == len(network['nodes'] * 2)+1
        for err in errors2[1:]:
            try:
                assert err.startswith("Unit mismatch")
            except AssertionError:
                assert err.startswith("Dimension mismatch")

    def test_type_compatibility(self, client, mock_template, mock_template_copy):
        """
            Check function that thests whether two types are compatible -- the
            overlapping attributes are specified with the same unit.
            Change the unit associated with an attribute for types in two idencical
            templates, and test. There should be 1 error returned.
            THen test comparison of identical types. No errors should return.
        """
        template_1 = mock_template
        template_2 = mock_template_copy

        diff_type_1_id = None
        same_type_1_id = None
        for typ in template_1.templatetypes:
            if typ.typeattrs:
                for ta in typ.typeattrs:
                    if ta.attr.name == 'node_attr_1':
                        diff_type_1_id = typ.id
                        ta.unit_id = client.get_unit_by_abbreviation("m^3").id
                    elif ta.attr.name == 'link_attr_1':
                        same_type_1_id = typ.id

        updated_template_1 = JSONObject(client.update_template(template_1))

        diff_type_2_id = None
        same_type_2_id = None
        for typ in template_2.templatetypes:
            if typ.typeattrs:
                for ta in typ.typeattrs:
                    if ta.attr.name == 'node_attr_1':
                        diff_type_2_id = typ.id
                        ta.unit_id = client.get_unit_by_abbreviation("cm^3").id
                    elif ta.attr.name == 'link_attr_1':
                        same_type_2_id = typ.id

        # Before updating template 2, check compatibility of types, where T1 has
        # a unit, but t2 does not.
        errors_diff = client.check_type_compatibility(diff_type_1_id, diff_type_2_id)
        assert len(errors_diff) == 1
        errors_same = client.check_type_compatibility(same_type_1_id, same_type_2_id)
        assert len(errors_same) == 0

        # Now update T2 so that the types have conflicting units.
        updated_template_2 = JSONObject(client.update_template(template_2))

        errors_diff = client.check_type_compatibility(diff_type_1_id, diff_type_2_id)
        assert len(errors_diff) == 1
        errors_same = client.check_type_compatibility(same_type_1_id, same_type_2_id)
        assert len(errors_same) == 0

        for typ in updated_template_1.templatetypes:
            if typ.typeattrs:
                for ta in typ.typeattrs:
                    if ta.attr.name == 'node_attr_1':
                        ta.unit_id = None

        # Update template 1 now so that it has no unit, but template 2 does.
        updated_template_1_a = client.update_template(updated_template_1)
        errors_diff = client.check_type_compatibility(diff_type_1_id, diff_type_2_id)
        assert len(errors_diff) == 1
        errors_same = client.check_type_compatibility(same_type_1_id, same_type_2_id)
        assert len(errors_same) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
"""
    NOTE:
    All the methods that has SKELETON key in the doc, are not yet implemented partially/fully

"""
# Example of a SKELETON METHOD
# def test_(self, client):
#     """
#         SKELETON
#     """
#     pass


import logging
import copy
import json
import hydra_base as hb
from hydra_base.exceptions import HydraError
import datetime
import pytest
from hydra_base.lib.objects import JSONObject
log = logging.getLogger(__name__)

class TestScopedAttribute:
    """
        Test for scoped-attribute-based functionality
    """

    def test_add_network_scoped_attribute(self, client, network_with_data):
        """
            Test adding a network-scoped attributes.
            1: Test adding a global attribute.
            2: Test adding a network-scoped attribute.
            3: Test that the network-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(network_id)
        """
        global_attr = JSONObject({
            "name": f'Global Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        network_scoped_attr = JSONObject({
            "name": f'Network Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "network_id": network_with_data.id
        })

        previous_all_attributes = client.get_attributes()

        client.add_attribute(global_attr)

        all_global_attributes = client.get_attributes()

        assert len(all_global_attributes) == len(previous_all_attributes) + 1

        #try add it again. SHould have no effect
        client.add_attribute(network_scoped_attr)

        global_attributes_no_network = client.get_attributes()
        network_scoped_attributes = client.get_attributes(network_id=network_with_data.id)

        #This should not have changed
        assert len(global_attributes_no_network) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new networks, plus the one we just added
        assert len(network_scoped_attributes) == 2


    def test_add_project_scoped_attribute(self, client, network_with_data):
        """
            Test adding a project-scoped attributes.
            1: Test adding a global attribute.
            2: Test adding a project-scoped attribute.
            3: Test that the project-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(project_id)
        """
        global_attr = JSONObject({
            "name": f'Global Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        project_scoped_attr = JSONObject({
            "name": f'Project Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "project_id": network_with_data.project_id
        })

        previous_all_attributes = client.get_attributes()

        client.add_attribute(global_attr)

        all_global_attributes = client.get_attributes()

        assert len(all_global_attributes) == len(previous_all_attributes) + 1

        #try add it again. SHould have no effect
        client.add_attribute(project_scoped_attr)

        global_attributes_no_project = client.get_attributes()
        project_scoped_attributes = client.get_attributes(project_id=network_with_data.project_id)

        #This should not have changed
        assert len(global_attributes_no_project) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new projects, plus the one we just added
        assert len(project_scoped_attributes) == 2


    def test_add_network_and_project_scoped_attribute(self, client, network_with_data):
        """
            Test adding a network-scoped attributes.
            1: Test adding a global attribute.
            2: Test adding a network-scoped attribute.
            3: Test that the network-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(network_id)
            4: Test adding a project-scoped attribute.
            5: Test that the project-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(project_id)
        """
        global_attr = JSONObject({
            "name": f'Global Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        network_scoped_attr = JSONObject({
            "name": f'Network Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "network_id": network_with_data.id
        })

        previous_all_attributes = client.get_attributes()

        client.add_attribute(global_attr)

        all_global_attributes = client.get_attributes()

        assert len(all_global_attributes) == len(previous_all_attributes) + 1

        #try add it again. SHould have no effect
        client.add_attribute(network_scoped_attr)

        global_attributes_no_network = client.get_attributes()
        network_scoped_attributes = client.get_attributes(network_id=network_with_data.id)

        #This should not have changed
        assert len(global_attributes_no_network) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(network_scoped_attributes) == 2


        project_scoped_attr = JSONObject({
            "name": f'Project Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "project_id": network_with_data.project_id
        })

        #try add it again. SHould have no effect
        client.add_attribute(project_scoped_attr)

        global_attributes_no_project = client.get_attributes()
        project_scoped_attributes = client.get_attributes(project_id=network_with_data.project_id)

        #This should not have changed
        assert len(global_attributes_no_project) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new projects, plus the one we just added
        assert len(project_scoped_attributes) == 2

        project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            network_id=network_with_data.id)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(project_and_network_scoped_attributes) == 4


        global_and_project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            network_id=network_with_data.id,
            include_global=True)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(global_and_project_and_network_scoped_attributes) == len(all_global_attributes) + 4

        #Now get project attributes, and include attributes from all networks within that project
        global_and_project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            include_global=True,
            include_network_attributes=True)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(global_and_project_and_network_scoped_attributes) == len(all_global_attributes) + 4


    def test_add_duplicate_scoped_attribute(self, client, network_with_data):
        """
            Test adding a scoped attribute which exists at a higher level.
        """
        global_attr = JSONObject({
            "name": f'Duplicate Attribute',
            "dimension_id": None
        })

        network_scoped_attr = JSONObject({
            "name": f'Duplicate Attribute',
            "dimension_id": None,
            "network_id": network_with_data.id
        })

        client.add_attribute(global_attr)

        with pytest.raises(hb.HydraError):
            client.add_attribute(network_scoped_attr)

    def test_add_attribute_to_project_and_network(self, client, network_with_data):
        """
            Test adding a scoped attribute which exists at a higher level.
        """
        dual_scoped_attr = JSONObject({
            "name": f'Duplicate Attribute',
            "dimension_id": None,
            "network_id": network_with_data.id,
            "project_id": network_with_data.project_id
        })

        with pytest.raises(hb.HydraError):
            client.add_attribute(dual_scoped_attr)

    def test_add_attribute_without_permission(self, client):
        """
            Test to make sure a non-admin cannot add a global attribute
        """
        global_scoped_attr = JSONObject({
            "name": f'Global Attribute',
            "dimension_id": None,
        })
        client.user_id = pytest.user_c.id
        with pytest.raises(PermissionError):
            client.add_attribute(global_scoped_attr)
        client.user_id = 1 # reset the user ID

    def test_add_network_scoped_attribute_without_ownership(self, client, network_with_data):
        """
            Test to ensure only the owner of a network can add ann attribute scoped
            to it.
        """
        network_scoped_attr = JSONObject({
            "name": f'Net Attribute',
            "dimension_id": None,
            "network_id" : network_with_data.id
        })
        user = client.testutils.create_user("UserD", role="developer")
        #User D does not own this network, so fail.
        client.user_id = user.id
        with pytest.raises(hb.PermissionError):
            client.add_attribute(network_scoped_attr)

        client.user_id = 1
        client.share_network(
            network_id=network_with_data.id,
            usernames=["UserC"],
            read_only=False,
            share=True)

        #User C is now an owner of the network, so it should succeedd
        client.user_id = pytest.user_c.id
        client.add_attribute(network_scoped_attr)
        client.user_id = 1 ## reset the user ID

    def test_add_project_scoped_attribute_without_ownership(self, client, network_with_data):
        """
            Test to ensure only the owner of a network can add ann attribute scoped
            to it.
        """
        project_scoped_attr = JSONObject({
            "name": f'Project Attribute',
            "dimension_id": None,
            "project_id" : network_with_data.project_id
        })
        user = client.testutils.create_user("UserD", role="developer")
        #User C does not own this network, so fail.
        client.user_id = user.id
        with pytest.raises(hb.PermissionError):
            client.add_attribute(project_scoped_attr)

        client.user_id = 1
        client.share_project(
            project_id=project_scoped_attr.project_id,
            usernames=["UserD"],
            read_only=False,
            share=True)

        #User C is now an owner of the network, so it should succeedd
        client.user_id = user.id
        client.add_attribute(project_scoped_attr)
        client.user_id = 1 ## reset the user ID

    def test_search_attribute_in_sub_project(self, client, projectmaker, networkmaker):
        """
           Test searching for an attribute defined on a parent project (p1) when searching from
           network n1. This requires recursively looking up the project tree to collate
           all attributes available to network n1. Attributes defined on P3 should not
           be visible.
                                 p1
                                /  \
                               p2   p3
                              /
                             p4
                            /
                           n1
        """
        client.user_id = 1 # force current user to be 1 to avoid potential inconsistencies
        proj_user = client.user_id
        proj1 = projectmaker.create(share=False)
        proj2 = projectmaker.create(share=False, parent_id=proj1.id)
        proj3 = projectmaker.create(share=False, parent_id=proj1.id)
        proj4 = projectmaker.create(share=False, parent_id=proj2.id)

        net1 = networkmaker.create(project_id=proj4.id)

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_project(proj4.id)

        with pytest.raises(HydraError):
            client.get_network(net1.id)

        #Now, as the main user, share P4 with user C
        client.user_id = proj_user
        client.share_network(net1.id, ['UserC'], False, False)

        client.add_attribute({'project_id': proj1.id,'name': 'p1_attr'})
        client.add_attribute({'project_id': proj3.id,'name': 'p3_attr'})
        client.add_attribute({'network_id': net1.id,'name': 'n1_attr'})

        matching_attributes = client.search_attributes('p1_', project_id=proj1.id)
        assert 'p1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('p1_', network_id=net1.id)
        assert 'p1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('n1_', network_id=net1.id)
        assert 'n1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('p3_', network_id=net1.id)
        assert 'p3_attr' not in [a.name for a in matching_attributes]

        #now do all the same things except with the shared user.
        client.user_id = pytest.user_c.id
        matching_attributes = client.search_attributes('p1_', project_id=proj1.id)
        assert 'p1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('p1_', network_id=net1.id)
        assert 'p1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('n1_', network_id=net1.id)
        assert 'n1_attr' in [a.name for a in matching_attributes]
        matching_attributes = client.search_attributes('p3_', network_id=net1.id)
        assert 'p3_attr' not in [a.name for a in matching_attributes]

        with pytest.raises(HydraError):
            client.search_attributes('p3_', project_id=proj3.id)



    def test_rescope_attribute(self, client, projectmaker, networkmaker):
        """
           Test to make sure that when a scoped attribute is added to a network
           which has a sibling containing the same attribute, then rather than
           adding a new attribute, the current attribute is re-scoped to the project
        """
        client.user_id = 1 # force current user to be 1 to avoid potential inconsistencies
        proj_user = client.user_id
        proj1 = projectmaker.create(share=False)

        proj1net1 = networkmaker.create(project_id=proj1.id)
        proj1net2 = networkmaker.create(project_id=proj1.id)

        #Add a child project which is at the same level as proj1net1/
        childproj = projectmaker.create(share=False, parent_id=proj1.id)
        childprojnet1 = networkmaker.create(project_id=childproj.id)


        proj2 = projectmaker.create(share=False)

        proj2net1 = networkmaker.create(project_id=proj2.id)

        net_scoped_attr = client.add_attribute({'network_id': proj1net1.id,'name': 'test_scoped_attr'})
        childproj_scoped_attr = client.add_attribute({'project_id': childproj.id,'name': 'test_scoped_attr'})
        #safety net to make sure it's been added properly
        net_scoped_attr = client.get_attribute_by_id(attr_id=net_scoped_attr.id)
        childproj_scoped_attr = client.get_attribute_by_id(attr_id=childproj_scoped_attr.id)

        #Make sure they've been addded properly
        assert  net_scoped_attr.network_id == proj1net1.id
        assert  childproj_scoped_attr.project_id == childproj.id

        #Add a resource attribute to both the child network and child project
        new_network_ra = client.add_resource_attribute('NODE', proj1net1.nodes[0].id, net_scoped_attr.id, is_var=False)
        new_child_project_ra = client.add_resource_attribute('NODE', childprojnet1.nodes[0].id, childproj_scoped_attr.id, is_var=False)

        #now add an attribute with the same name at the project level, thus creating a potential
        #duplication
        newly_scoped_attr = client.add_attribute({'project_id': proj1.id,'name': 'test_scoped_attr'})

        with pytest.raises(HydraError):
            client.get_attribute_by_id(attr_id=net_scoped_attr.id)

        with pytest.raises(HydraError):
            _ = client.get_attribute_by_id(attr_id=childproj_scoped_attr.id)

        #check it has been rescoped
        assert newly_scoped_attr.project_id == proj1.id

        updated_network_ra = client.get_resource_attribute(new_network_ra.id)
        updated_child_project_ra = client.get_resource_attribute(new_child_project_ra.id)

        assert updated_network_ra.attr_id == newly_scoped_attr.id
        assert updated_child_project_ra.attr_id == newly_scoped_attr.id

        matching_attributes = client.search_attributes('test_scoped', network_id=proj1net1.id)

        assert len(matching_attributes) == 1 # the default scoped attrs plus this one.

        assert 'test_scoped_attr' in [a.name for a in matching_attributes]

        assert matching_attributes[0].id == newly_scoped_attr.id

        proj2_net_scoped_attr_ = client.add_attribute({'network_id': proj2net1.id,'name': 'test_scoped_attr'})
        #safety net to make sure it's been added properly
        proj2_net_scoped_attr = client.get_attribute_by_id(attr_id=proj2_net_scoped_attr_.id)
        new_ra = client.add_resource_attribute('NODE', proj2net1.nodes[0].id, proj2_net_scoped_attr.id, is_var=False)
        assert  proj2_net_scoped_attr.network_id == proj2net1.id

        #now add an attribute with the same name at the project level, thus creating a potential
        #duplication
        proj2_newly_scoped_attr = client.add_attribute({'project_id': proj2.id,'name': 'test_scoped_attr'})

        with pytest.raises(HydraError):
            _ = client.get_attribute_by_id(attr_id=proj2_net_scoped_attr.id)

        #check it has been rescoped
        assert proj2_newly_scoped_attr.project_id == proj2.id

        proj2_scoped_attr = client.get_attribute_by_id(attr_id=proj2_newly_scoped_attr.id)
        assert proj2_scoped_attr != newly_scoped_attr

    def test_different_projects_same_name_attributes(self, client, projectmaker, networkmaker):
        """
            Verifies that attribute deletion following rescoping occurs only within the
            correct bounds.
            Two same-name-and-dimension attrs added in peer projects should exist
            independently; the addition of the second should have no consequences for the first.
        """
        client.user_id = 1
        proj_user = client.user_id
        proj1 = projectmaker.create(name="Project 1", share=False)
        proj2 = projectmaker.create(name="Project 2", share=False)

        project1_scoped_attr = JSONObject({
            "name": f"Project Attribute",
            "dimension_id": None,
            "project_id": proj1.id
        })

        project2_scoped_attr = JSONObject({
            "name": f"Project Attribute",
            "dimension_id": None,
            "project_id": proj2.id
        })

        proj1_attr = client.add_attribute(project1_scoped_attr)
        net1 = networkmaker.create(project_id=proj1.id)
        proj1_ra = client.add_resource_attribute('NODE', net1.nodes[0].id, proj1_attr.id, is_var=False)
        proj1_attrs_before = client.get_attributes(project_id=proj1.id)

        proj2_attr = client.add_attribute(project2_scoped_attr)
        net2 = networkmaker.create(project_id=proj2.id)
        proj2_ra = client.add_resource_attribute('NODE', net2.nodes[0].id, proj2_attr.id, is_var=False)
        proj1_attrs_after = client.get_attributes(project_id=proj1.id)

        assert len(proj1_attrs_before) == len(proj1_attrs_after)
        assert proj1_attrs_before == proj1_attrs_after


    def test_bulk_add_network_and_project_scoped_attribute(self, client, network_with_data):
        """
            Test adding a network-scoped attributes.
            1: Test adding a global attribute.
            2: Test adding a network-scoped attribute.
            3: Test that the network-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(network_id)
            4: Test adding a project-scoped attribute.
            5: Test that the project-scoped attribute doesn't appear in the
            get_attributes() function but does appear in the get_attributes(project_id)
        """
        global_attr = JSONObject({
            "name": f'Global Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        network_scoped_attr = JSONObject({
            "name": f'Network Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "network_id": network_with_data.id
        })

        project_scoped_attr = JSONObject({
            "name": f'Project Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "project_id": network_with_data.project_id
        })


        previous_all_attributes = client.get_attributes()

        new_attributes = client.add_attributes([global_attr,
                                                network_scoped_attr,
                                                project_scoped_attr])

        all_global_attributes = client.get_attributes()

        assert len(all_global_attributes) == len(previous_all_attributes) + 1

        #try add it again. SHould have no effect
        client.add_attributes([global_attr,
                               network_scoped_attr,
                               project_scoped_attr])

        global_attributes_no_network = client.get_attributes()
        network_scoped_attributes = client.get_attributes(
            network_id=network_with_data.id)

        #This should not have changed
        assert len(global_attributes_no_network) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new
        #networks and projects, plus the ones we just added
        assert len(network_scoped_attributes) == 2

        #try add it again. SHould have no effect
        client.add_attribute(project_scoped_attr)

        global_attributes_no_project = client.get_attributes()
        project_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id)

        #This should not have changed
        assert len(global_attributes_no_project) == len(all_global_attributes)

        #It's 2 because there is one added by default to all new projects, plus the one we just added
        assert len(project_scoped_attributes) == 2

        project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            network_id=network_with_data.id)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(project_and_network_scoped_attributes) == 4


        global_and_project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            network_id=network_with_data.id,
            include_global=True)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(global_and_project_and_network_scoped_attributes) == len(all_global_attributes) + 4

        #Now get project attributes, and include attributes from all networks within that project
        global_and_project_and_network_scoped_attributes = client.get_attributes(
            project_id=network_with_data.project_id,
            include_global=True,
            include_network_attributes=True)

        #It's 4 because there is one added by default to all new networks and projects, plus the ones we just added
        assert len(global_and_project_and_network_scoped_attributes) == len(all_global_attributes) + 4


    def test_add_network_scoped_attribute_with_attribute_already_existing(self, client, network_with_data):
        """
            Test that when adding an attribute to a network, where that attribute is already
            scoped higher, then return the higher scoped attribute instead of failing
        """
        global_attr = JSONObject({
            "name": f'Global Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        network_scoped_attr = JSONObject({
            "name": f'Network Attribute {datetime.datetime.now()}',
            "dimension_id": None,
            "network_id": network_with_data.id
        })

        project_attr_name =  f'Project Attribute {datetime.datetime.now()}'

        project_scoped_attr = JSONObject({
            "name": project_attr_name,
            "dimension_id": None,
            "project_id": network_with_data.project_id
        })


        previous_all_attributes = client.get_attributes()

        new_attributes = client.add_attributes([global_attr, network_scoped_attr, project_scoped_attr])

        new_project_attr = list(filter(lambda x:x.project_id==network_with_data.project_id, new_attributes))[0]

        scoped_ids = [a.id for a in new_attributes]

        #Now add the attribute which is scoped to the project, to the network
        duplicate_network_scoped_attr = JSONObject({
            "name": project_attr_name,
            "dimension_id": None,
            "network_id": network_with_data.id
        })


        new_attributes = client.add_attributes([duplicate_network_scoped_attr])

        #should return an existing scoped attribute (the project one)
        assert len(new_attributes) == 1
        assert new_attributes[0].id == new_project_attr.id

        new_attribute = client.add_attribute(duplicate_network_scoped_attr)

        assert new_attribute.id == new_project_attr.id



import logging

from bson.objectid import ObjectId
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from hydra_base import config

log = logging.getLogger(__name__)


class MongoStorageAdapter():
    """
    Provides an interface to DatasetManager instances to access MongoDB storage
    """
    def __init__(self):
        mongo_config = self.__class__.get_mongo_config()
        host = mongo_config["host"]
        port = mongo_config["port"]
        user = mongo_config["user"]
        passwd = mongo_config["passwd"]
        self.db_name = mongo_config["db_name"]
        self.datasets = mongo_config["datasets"]

        """ Mongo usernames/passwds require percent encoding of `:/?#[]@` chars """
        user, passwd = percent_encode(user), percent_encode(passwd)
        authtext = f"{user}:{passwd}@" if (user and passwd) else ""
        conntext = f"mongodb://{authtext}{host}:{port}/{self.db_name}"
        try:
            self.client = MongoClient(conntext)
        except pymongo.errors.ServerSelectionTimeoutError as sste:
            log.critical(f"Unable to connect to Mongo server {conntext}: {sste}")
            raise sste

        self.db = self.client[self.db_name]

    @staticmethod
    def get_mongo_config(config_key="mongodb"):
        numeric = ("threshold",)
        mongo_keys = [k for k in config.CONFIG.options(config_key) if k not in config.CONFIG.defaults()]
        mongo_items = {k: config.CONFIG.get(config_key, k) for k in mongo_keys}
        for k in numeric:
            mongo_items[k] = int(mongo_items[k])

        return mongo_items

    def __del__(self):
        """ Close connection on object destruction """
        if hasattr(self, "client"):
            self.client.close()

    def get_document_by_object_id(self, object_id: str, collection=None):
        """ Retrieve the document with the specified object_id from a collection """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        doc = path.find_one({"_id": ObjectId(object_id)})
        return doc

    def get_document_by_oid_inst(self, object_id: ObjectId, collection=None):
        """ Retrieve the document with the specified object_id from a collection """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        doc = path.find_one({"_id": object_id})
        return doc

    def delete_document_by_object_id(self, object_id: str, collection=None):
        """ Delete the document with the specified object_id from a collection """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        doc = {"_id": ObjectId(object_id)}
        path.delete_one(doc)

    def set_document_value(self, object_id: str, value, collection=None):
        """
        Set the `value` key of the document with the specified object_id
        to the `value` argument
        """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        doc = {"_id": ObjectId(object_id)}
        path.update_one(doc, {"$set": {"value": value}})

    def insert_document(self, value, collection=None):
        """ Insert a document with the specified `value` into a collection """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        result = path.insert_one({"value": value})
        return result.inserted_id

    def bulk_insert_values(self, values, collection=None):
        """
        Insert a list of `values` into a collection.
        Returns an instance of InsertManyResults containing an .inserted_ids list
        """
        collection = collection if collection else self.datasets
        path = self.db[collection]
        data = [{"value": value} for value in values]
        inserted = path.insert_many(data)
        return inserted

    @property
    def default_collection(self):
        return self.datasets




def percent_encode(s, xchars=":/?#[]@"):
    return "".join(f"%{ord(char):2X}" if char in xchars else char for char in s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import time
from ..base import *

from ..dataset import Dataset, Metadata
from ..network import ResourceAttr
from ..attributes import Attr
from ..ownership import DatasetOwner
from .resourcegroupitem import ResourceGroupItem

__all__ = ['Scenario', 'ResourceScenario']

class ResourceScenario(Base, Inspect):
    """
    """

    __tablename__='tResourceScenario'

    dataset_id = Column(Integer(), ForeignKey('tDataset.id'), nullable=False)
    scenario_id = Column(Integer(), ForeignKey('tScenario.id'), primary_key=True, nullable=False, index=True)
    resource_attr_id = Column(Integer(), ForeignKey('tResourceAttr.id'), primary_key=True, nullable=False, index=True)
    source           = Column(String(60))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    dataset      = relationship('Dataset', backref=backref("resourcescenarios", order_by=dataset_id))
    scenario     = relationship('Scenario', backref=backref("resourcescenarios", order_by=scenario_id, cascade="all, delete-orphan"))
    resourceattr = relationship('ResourceAttr', backref=backref("resourcescenarios", cascade="all, delete-orphan"), uselist=False)

    _parents  = ['tScenario', 'tResourceAttr']
    _children = ['tDataset']

    def get_dataset(self, user_id):
        dataset = get_session().query(Dataset.id,
                Dataset.type,
                Dataset.unit_id,
                Dataset.name,
                Dataset.hidden,
                case([(and_(Dataset.hidden=='Y', DatasetOwner.user_id is not None), None)],
                        else_=Dataset.value).label('value'),
                case([(and_(Dataset.hidden=='Y', DatasetOwner.user_id is not None), None)],
                        else_=Dataset.value).label('value')).filter(
                Dataset.id==self.id).outerjoin(DatasetOwner,
                                    and_(Dataset.id==DatasetOwner.dataset_id,
                                    DatasetOwner.user_id==user_id)).one()

        return dataset

    @property
    def value(self):
        return self.dataset

class Scenario(Base, Inspect):
    """
    """

    __tablename__='tScenario'
    __table_args__ = (
        UniqueConstraint('network_id', 'name', name="unique scenario name"),
    )

    id = Column(Integer(), primary_key=True, index=True, nullable=False)
    name = Column(String(200),  nullable=False)
    description = Column(String(1000))
    layout  = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'),  nullable=True)
    status = Column(String(1),  nullable=False, server_default=text(u"'A'"))
    network_id = Column(Integer(), ForeignKey('tNetwork.id'), index=True)
    start_time = Column(String(60))
    end_time = Column(String(60))
    locked = Column(String(1),  nullable=False, server_default=text(u"'N'"))
    time_step = Column(String(60))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    created_by = Column(Integer(), ForeignKey('tUser.id'), nullable=False)
    parent_id = Column(Integer(), ForeignKey('tScenario.id'), nullable=True)

    network = relationship('Network', backref=backref("scenarios", order_by=id))
    parent = relationship('Scenario', remote_side=[id],
        backref=backref("children", order_by=id))

    _parents  = ['tNetwork']
    _children = ['tResourceScenario']

    def add_resource_scenario(self, resource_attr, dataset=None, source=None):
        rs_i = ResourceScenario()
        if resource_attr.id is None:
            rs_i.resourceattr = resource_attr
        else:
            rs_i.resource_attr_id = resource_attr.id

        if dataset.id is None:
            rs_i.dataset = dataset
        else:
            rs_i.dataset_id = dataset.id
        rs_i.source = source
        self.resourcescenarios.append(rs_i)

    def add_resourcegroup_item(self, ref_key, resource, group_id):
        group_item_i = ResourceGroupItem()
        group_item_i.group_id = group_id
        group_item_i.ref_key  = ref_key
        if ref_key == 'GROUP':
            group_item_i.subgroup = resource
        elif ref_key == 'NODE':
            group_item_i.node     = resource
        elif ref_key == 'LINK':
            group_item_i.link     = resource
        self.resourcegroupitems.append(group_item_i)

    def get_data(self, user_id,
        child_data=None,
        get_parent_data=False,
        ra_ids=None,
        include_results=True,
        include_only_results=False,
        include_data_types=None,
        exclude_data_types=None,
        include_values=True,
        include_data_type_values=None,
        exclude_data_type_values=None,
        include_metadata=True):
        """
            Return all the resourcescenarios relevant to this scenario.
            If this scenario inherits from another, look up the tree to compile
            an exhaustive list of resourcescnearios, removing any duplicates, prioritising
            the ones closest to this scenario (my immediate parent's values are used instead
            of its parents)

            If an explicit list of RAs is provided, only return data for these. This is used
            when requesting data for a specific resource, for example.
        """

        #This avoids python's mutable keyword argumets causing child_data to keep its values beween
        #function calls
        if child_data is None:
            child_data = []

        #Idenify all existing resource attr ids, which take priority over anything in here
        childrens_ras = []
        for child_rs in child_data:
            childrens_ras.append(child_rs.resource_attr_id)
        t = time.time()
        resourcescenarios = self.get_all_resourcescenarios(
            user_id,
            ra_ids=ra_ids,
            include_results=include_results,
            include_only_results=include_only_results,
            include_metadata=include_metadata,
            include_data_types=include_data_types,
            exclude_data_types=exclude_data_types,
            include_values=include_values,
            include_data_type_values=include_data_type_values,
            exclude_data_type_values=exclude_data_type_values
            )
        log.info(f"get_all_resourcescenarios took {time.time() - t:.2f} seconds")
        for this_rs in resourcescenarios:
            if this_rs.resource_attr_id not in childrens_ras:
                child_data.append(this_rs)

        if self.parent is not None and get_parent_data is True:
            return self.parent.get_data(user_id, child_data=child_data,
                                        get_parent_data=get_parent_data,
                                       ra_ids=ra_ids)

        return child_data

    def get_all_resourcescenarios(self, user_id,
        ra_ids=None,
        include_results=True,
        include_only_results=False,
        include_data_types=None,
        exclude_data_types=None,
        include_values=True,
        include_data_type_values=None,
        exclude_data_type_values=None,
        include_metadata=True):
        """
            Get all the resource scenarios in a network, across all scenarios
            returns a dictionary of dict objects, keyed on scenario_id
        """
        log.debug("Starting dataset query")

        rs_qry = get_session().query(
                    Dataset.type,
                    Dataset.unit_id,
                    Dataset.name,
                    Dataset.hash,
                    Dataset.cr_date,
                    Dataset.created_by,
                    Dataset.hidden,
                    ResourceScenario.dataset_id,
                    ResourceScenario.scenario_id,
                    ResourceScenario.resource_attr_id,
                    ResourceScenario.source,
                    ResourceAttr.attr_id,
                    ResourceAttr.attr_is_var,
                    ResourceAttr.ref_key,
                    ResourceAttr.node_id,
                    ResourceAttr.network_id,
                    ResourceAttr.link_id,
                    ResourceAttr.group_id,
                    Attr.name.label('attr_name'),
                    Attr.description.label('attr_description')
        ).outerjoin(DatasetOwner, and_(DatasetOwner.dataset_id==Dataset.id, DatasetOwner.user_id==user_id)).filter(
                    or_(Dataset.hidden=='N', Dataset.created_by==user_id, DatasetOwner.user_id != None),
                    ResourceAttr.id == ResourceScenario.resource_attr_id,
                    Scenario.id==ResourceScenario.scenario_id,
                    Scenario.id==self.id,
                    Dataset.id==ResourceScenario.dataset_id,
                    Attr.id==ResourceAttr.attr_id)

        if include_results is False:
            rs_qry = rs_qry.filter(ResourceAttr.attr_is_var=='N')

        if include_only_results is True:
            rs_qry = rs_qry.filter(ResourceAttr.attr_is_var=='Y')

        if ra_ids is not None:
            rs_qry = rs_qry.filter(ResourceScenario.resource_attr_id.in_(ra_ids))
        
        if include_data_types is not None:
            rs_qry = rs_qry.filter(func.upper(Dataset.type).in_([t.upper() for t in include_data_types]))
        
        if exclude_data_types is not None:
            rs_qry = rs_qry.filter(func.upper(Dataset.type).not_in([t.upper() for t in exclude_data_types]))

        excl_values_rs = []
        if include_values is True:
            if include_data_type_values is not None:
                rs_qry = rs_qry.filter(func.upper(Dataset.type).in_([t.upper() for t in include_data_type_values]))

            if exclude_data_type_values is not None:
                excl_values_qry = rs_qry.filter(func.upper(Dataset.type).in_([t.upper() for t in exclude_data_type_values]))
                rs_qry = rs_qry.filter(func.upper(Dataset.type).not_in([t.upper() for t in exclude_data_type_values]))
                excl_values_rs = excl_values_qry.all()

            rs_qry = rs_qry.add_columns(Dataset.value)

        non_dataframe_rs = rs_qry.all()
        
        all_rs = non_dataframe_rs + excl_values_rs

        log.info(f"Ending dataset query -- {len(all_rs)} results")
        t = time.time()
        processed_rs = []
        for rs in all_rs:
            rs_obj = JSONObject({
                'resource_attr_id': rs.resource_attr_id,
                'scenario_id':rs.scenario_id,
                'dataset_id':rs.dataset_id,
                'resourceattr': {
                    'id': rs.resource_attr_id,
                    'attr_id':rs.attr_id,
                    'attr_is_var': rs.attr_is_var,
                    'ref_key': rs.ref_key,
                    'node_id': rs.node_id,
                    'link_id': rs.link_id,
                    'network_id': rs.network_id,
                    'group_id': rs.group_id,
                    'attr':{
                        'name': rs.attr_name,
                        'description':rs.attr_description,
                        'id': rs.attr_id
                    }
                },
                'dataset': {
                    'id':rs.dataset_id,
                    'type' : rs.type,
                    'unit_id' : rs.unit_id,
                    'name' : rs.name,
                    'hash' : rs.hash,
                    'cr_date':rs.cr_date,
                    'created_by':rs.created_by,
                    'hidden':rs.hidden,
                    'value': getattr(rs, 'value', None),
                    'metadata':{},
                }
            }, normalize=False)

            processed_rs.append(rs_obj)
        log.info(f"Datasets processed in {time.time() - t:.2f} seconds")

        ## If metadata is requested, use a dedicated query to extract metadata
        ## from the scenario's datasets,
        ## and enter them into a lookup table, keyed by dataset_id so they can
        ## be extracted later.
        metadata_lookup = {}
        if include_metadata is True:
            dataset_ids = [rs.dataset.id for rs in processed_rs]
            metadata = get_session().query(Metadata)\
                        .join(Dataset)\
                        .join(ResourceScenario)\
                        .filter(ResourceScenario.scenario_id == self.id).all()
            for m in metadata:
                if metadata_lookup.get(m.dataset_id):
                    metadata_lookup[m.dataset_id][m.key] = m.value
                else:
                    metadata_lookup[m.dataset_id] = {m.key:m.value}
            for rs in processed_rs:
                if rs.dataset.value is not None:
                    rs.dataset.metadata = metadata_lookup.get(rs.dataset.id, {})
        return processed_rs

    def get_group_items(self, child_items=None, get_parent_items=False):
        """
            Return all the resource group items relevant to this scenario.
            If this scenario inherits from another, look up the tree to compile
            an exhaustive list of resource group items, removing any duplicates, prioritising
            the ones closest to this scenario (my immediate parent's values are used instead
            of its parents)
        """

        #This avoids python's mutable keyword argumets causing child_items to keep its values beween
        #function calls
        if child_items == None:
            child_items = []

        #Idenify all existing resource attr ids, which take priority over anything in here
        childrens_groups = []
        for child_rgi in child_items:
            childrens_groups.append(child_rgi.group_id)

        #Add resource attributes which are not defined already
        for this_rgi in self.resourcegroupitems:
            if this_rgi.group_id not in childrens_groups:
                child_items.append(this_rgi)

        if self.parent is not None and get_parent_items is True:
            return self.parent.get_group_items(child_items=child_items,
                                               get_parent_items=get_parent_items)

        return child_items







#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from . import *


__all__ = ['Attr', 'AttrMap', 'AttrGroup', 'AttrGroupItem']

class Attr(Base, Inspect):
    """
    An Attribute Definition
    """

    __tablename__='tAttr'

    __table_args__ = (
        UniqueConstraint('name', 'dimension_id', 'network_id', 'project_id', name="unique name dimension_id"),
    )

    id = Column(Integer(), primary_key=True, nullable=False)
    name = Column(String(200),  nullable=False)
    dimension_id = Column(Integer(), ForeignKey('tDimension.id'), nullable=True)
    description = Column(String(1000))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    project_id = Column(Integer(), ForeignKey('tProject.id'), nullable=True)
    network_id = Column(Integer(), ForeignKey('tNetwork.id'), nullable=True)

    network = relationship('Network', foreign_keys=[network_id], backref=backref('scopedattributes', uselist=True, cascade="all, delete-orphan"), lazy='joined')
    project = relationship('Project', foreign_keys=[project_id], backref=backref('scopedattributes', uselist=True, cascade="all, delete-orphan"), lazy='joined')
    dimension = relationship('Dimension', foreign_keys=[dimension_id], backref=backref("attributes", uselist=True))

    _parents = ['tDimension']
    _children = []

class AttrMap(Base, Inspect):
    """
    """

    __tablename__='tAttrMap'

    attr_id_a = Column(Integer(), ForeignKey('tAttr.id'), primary_key=True, nullable=False)
    attr_id_b = Column(Integer(), ForeignKey('tAttr.id'), primary_key=True, nullable=False)

    attr_a = relationship("Attr", foreign_keys=[attr_id_a], backref=backref('maps_to', order_by=attr_id_a))
    attr_b = relationship("Attr", foreign_keys=[attr_id_b], backref=backref('maps_from', order_by=attr_id_b))

class AttrGroup(Base, Inspect):

    """
        **exclusive** : If 'Y' then an attribute in this group cannot be in any other groups

    """

    __tablename__='tAttrGroup'

    __table_args__ = (
        UniqueConstraint('name', 'project_id', name="unique attr group name"),
    )

    id               = Column(Integer(), primary_key=True, nullable=False, index=True)
    name             = Column(String(200), nullable=False)
    description      = Column(String(1000))
    layout           = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'),  nullable=True)
    exclusive        = Column(String(1),  nullable=False, server_default=text(u"'N'"))
    project_id       = Column(Integer(), ForeignKey('tProject.id'), primary_key=False, nullable=False)
    cr_date          = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    project          = relationship('Project', backref=backref('attrgroups', uselist=True, cascade="all, delete-orphan"), lazy='joined')


    _parents  = ['tProject']
    _children = []

class AttrGroupItem(Base, Inspect):
    """
        Items within an attribute group. Groupings are network dependent, and you can't
        have an attribute in a group twice, or an attribute in two groups.
    """

    __tablename__='tAttrGroupItem'

    group_id    = Column(Integer(), ForeignKey('tAttrGroup.id'), primary_key=True, nullable=False)
    attr_id    = Column(Integer(), ForeignKey('tAttr.id'), primary_key=True, nullable=False)
    network_id    = Column(Integer(), ForeignKey('tNetwork.id'), primary_key=True, nullable=False)

    group = relationship('AttrGroup', backref=backref('items', uselist=True, cascade="all, delete-orphan"), lazy='joined')
    attr = relationship('Attr')
    network = relationship('Network', backref=backref('attrgroupitems', uselist=True, cascade="all, delete-orphan"), lazy='joined')


    _parents  = ['tAttrGroup']
    _children = []


"""rule

Revision ID: 56ea2e814437
Revises: 6b24d6443e31
Create Date: 2018-04-17 16:01:19.297928

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '56ea2e814437'
down_revision = '6b24d6443e31'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tRule

        try:
            op.alter_column('tRule', 'rule_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tRule', 'rule_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tRule', 'rule_description', new_column_name='description', existing_type=sa.String(1000))
            op.alter_column('tRule', 'value', new_column_name='value', existing_type=sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'), nullable=True)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tRule
        try:
            op.drop_table('tRule_new')
        except:
            log.info("tRule_new isn't there")

        try:
            # ## tRule
            op.create_table(
                'tRule_new',
                    sa.Column('id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('name',sa.String(60), nullable=False),
                    sa.Column('description',sa.String(1000), nullable=False),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.Column('ref_key',sa.String(60),  nullable=False, index=True),
                    sa.Column('value',sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'),  nullable=True),
                    sa.Column('status',sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                    sa.Column('scenario_id',sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False),
                    sa.Column('network_id',sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                    sa.Column('node_id',sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                    sa.Column('link_id',sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                    sa.Column('group_id',sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
                    sa.UniqueConstraint('scenario_id', 'name', name="unique rule name")
            )

            op.execute("insert into tRule_new (id, name, description, ref_key, status, value, scenario_id, network_id, node_id, link_id, group_id, cr_date) select rule_id, rule_name, rule_description, ref_key, status, value, scenario_id, network_id, node_id, link_id, group_id, cr_date from tRule")

            op.rename_table('tRule','tRule_old')
            op.rename_table('tRule_new', 'tRule')
            op.drop_table('tRule_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tRule

        try:
            op.alter_column('tRule', 'id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tRule', 'name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tRule', 'description', new_column_name='description', existing_type=sa.String(1000))
            op.alter_column('tRule', 'value', new_column_name='value', existing_type=sa.LargeBinary(), nullable=True)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tRule
        try:
            op.drop_table('tRule_new')
        except:
            log.info("tRule_new isn't there")

        try:
            # ## tRule
            op.create_table(
                'tRule_new',
                    sa.Column('rule_id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('rule_name',sa.String(60), nullable=False),
                    sa.Column('rule_description',sa.String(1000), nullable=False),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.Column('ref_key',sa.String(60),  nullable=False, index=True),
                    sa.Column('value',sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'),  nullable=True),
                    sa.Column('status',sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                    sa.Column('scenario_id',sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False),
                    sa.Column('network_id',sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                    sa.Column('node_id',sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                    sa.Column('link_id',sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                    sa.Column('group_id',sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
                    sa.UniqueConstraint('scenario_id', 'name', name="unique rule name")
            )

            op.execute("insert into tRule_new (rule_id, rule_name, rule_description, ref_key, status, value, scenario_id, network_id, node_id, link_id, group_id, cr_date) select id, name, description, ref_key, status, value, scenario_id, network_id, node_id, link_id, group_id, cr_date from tRule")

            op.rename_table('tRule','tRule_old')
            op.rename_table('tRule_new', 'tRule')
            op.drop_table('tRule_old')

        except Exception as e:
            log.exception(e)



"""change password type

Revision ID: 6883f6e39176
Revises: 21e8d704c020
Create Date: 2018-08-08 12:06:59.579756

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6883f6e39176'
down_revision = '21e8d704c020'
branch_labels = None
depends_on = None

import logging
log = logging.getLogger(__name__)

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tUser

        try:
            op.alter_column('tUser', 'password', existing_type=sa.LargeBinary(), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tUser_new')
        except:
            log.info('tScenario_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tUser_new',
                sa.Column('id', sa.Integer(), primary_key=True),
                sa.Column('username', sa.Text(200), nullable=False),
                sa.Column('password', sa.LargeBinary),
                sa.Column('display_name', sa.String(60),  nullable=False, server_default=sa.text(u"''")),
                sa.Column('last_login', sa.TIMESTAMP()),
                sa.Column('last_edit', sa.TIMESTAMP()),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP'))
            )

            op.execute("insert into tUser_new (id, username, password, display_name, last_login, last_edit, cr_date) select id, username, password, display_name, last_login, last_edit, cr_date from tUser")

            op.rename_table('tUser','tUser_old')
            op.rename_table('tUser_new', 'tUser')
            op.drop_table('tUser_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tUser

        try:
            op.alter_column('tUser', 'password', existing_type=sa.Text(), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tUser_new')
        except:
            log.info('tScenario_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tUser_new',
                sa.Column('id', sa.Integer(), primary_key=True),
                sa.Column('username', sa.Text(200), nullable=False),
                sa.Column('password', sa.Text(1000)),
                sa.Column('display_name', sa.String(60),  nullable=False, server_default=sa.text(u"''")),
                sa.Column('last_login', sa.TIMESTAMP()),
                sa.Column('last_edit', sa.TIMESTAMP()),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP'))
            )

            op.execute("insert into tUser_new (id, username, password, display_name, last_login, last_edit, cr_date) select id, username, password, display_name, last_login, last_edit, cr_date from tUser")

            op.rename_table('tUser','tUser_old')
            op.rename_table('tUser_new', 'tUser')
            op.drop_table('tUser_old')

        except Exception as e:
            log.exception(e)



# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2020 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

"""
    This module contains all functions which relate to the application,
    validation or manipulation of resources (nodes, links, groups, networks)
    in relation to a template, rather than manipulating
    the templates themselves.
    These include applying types to resources, validating resources etc.
"""

import json
import logging

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import noload, joinedload
from sqlalchemy import or_, and_

from hydra_base import db
from hydra_base.db.model import Template, TemplateType, Attr, \
                                Network, Node, Link, ResourceGroup,\
                                ResourceType, ResourceAttr, ResourceScenario, Scenario
from hydra_base.lib.objects import JSONObject
from hydra_base.exceptions import HydraError
from hydra_base.util import dataset_util
from hydra_base.lib import units
from hydra_base.util.permissions import required_perms

log = logging.getLogger(__name__)

def get_network_template(network_id, reference_type_id):
    """
        If a network type is defined with a child template id (it has been created
        using a child template), then find this so that it can be used when
        inserting nodes & links into that template.

        This is to avoid having to set the child template ID on all add_node add_link
        functions etc.

        The reference_type_id is the type of the node or link being inserted, and
        is needed because a network may have multiple types, and so a type from the same
        template as one of those network types needs to be used for reference.
    """
    #Get all the network types
    network_types = db.DBSession.query(ResourceType)\
        .filter(ResourceType.network_id == network_id).all()

    #get the template type of the incoming resource
    ref_type = db.DBSession.query(TemplateType)\
        .filter(TemplateType.id == reference_type_id).one()

    #a bit roundabout, but this gets the fully-populated, inherited child type
    ref_type_inherited = ref_type.template.get_type(reference_type_id)

    #Now go through each network type and try to find the matching template.
    #Assume that a network cannot have 2 types which inherit from the same template.
    for nt in network_types:
        network_tt = nt.get_templatetype()

        if ref_type.template_id == network_tt.template_id:
            #if the network's type and the incoming type's template ID, we've found
            #a match.
            return nt.child_template_id
        elif ref_type.template_id == nt.child_template_id:
            #If the child template id of the network is the same as the template ID of the incoming node type
            #then we've got a match -- this implies that the node type is defined in the child
            #itself, meaning it has been modified compared to its parent in some way.
            return nt.child_template_id
        #Get the ID of the template which is the parent of the child template which
        #the network was created with
        #TODO: THis only works for single inheritance. We can only look up one
        #level in the inheritance tree.
        network_template_parent_id = db.DBSession.query(Template)\
            .filter(Template.id==network_tt.template_id).one().parent_id

        if ref_type.template_id == network_template_parent_id:
            #if the parent template of the network type is the same as
            #the the template ID of this type
            return nt.child_template_id

    return None

def _get_type(type_id):
    """
        Utility function to get a template type by querying the parent for it.
        This must be done because types are constructed using template inheritance,
        so cannot be queried directly
    """

    log.info("Getting type %s", type_id)

    type_i = db.DBSession.query(TemplateType).filter(TemplateType.id == type_id).one()

    if type_i.parent_id is None:
        return type_i

    template_i = db.DBSession.query(Template)\
            .filter(Template.id == type_i.template_id).one()

    type_i = template_i.get_type(type_id)
    type_i.template = JSONObject(template_i)

    return type_i

@required_perms('get_template')
def get_types_by_attr(resource, resource_type, template_id=None, **kwargs):
    """
        Using the attributes of the resource, get all the
        types that this resource matches.
        args:
            resource (a resource object (node, link etc), assumed to have a
                     '.attributes' attribute)
            template_id: The ID of a template, which will filter the result to
                         just types in that template
        returns:
            dict: keyed on the template name, with the
            value being the list of type names which match the resources
            attributes.
    """

    resource_type_templates = []

    #Create a list of all of this resources attributes.
    attr_ids = []
    for res_attr in resource.attributes:
        attr_ids.append(res_attr.attr_id)
    all_resource_attr_ids = set(attr_ids)

    all_templates_qry = db.DBSession.query(Template)

    if template_id is not None:
        all_templates_qry = all_templates_qry.filter(Template.id == template_id)

    all_templates = all_templates_qry.all()

    all_types = []
    for template in all_templates:
        template_types = template.get_types()
        all_types.extend(template_types)

    #tmpl type attrs must be a subset of the resource's attrs
    for ttype in all_types:

        if ttype.resource_type != resource_type:
            continue

        type_attr_ids = []
        for typeattr in ttype.typeattrs:
            type_attr_ids.append(typeattr.attr_id)
        if set(type_attr_ids).issubset(all_resource_attr_ids):
            resource_type_templates.append(ttype)

    return resource_type_templates



@required_perms("edit_network")
def apply_template_to_network(template_id, network_id, **kwargs):
    """
        For each node and link in a network, check whether it matches
        a type in a given template. If so, assign the type to the node / link.
    """

    net_i = db.DBSession.query(Network).filter(Network.id == network_id).one()
    #There should only ever be one matching type, but if there are more,
    #all we can do is pick the first one.
    try:
        network_template_i = db.DBSession.query(Template).filter(
            Template.id == template_id).one()

        for template_type in network_template_i.get_types():
            if template_type.resource_type == 'NETWORK':
                network_type_id = template_type.id
                assign_type_to_resource(network_type_id, 'NETWORK', network_id, **kwargs)
                break
        else:
            raise NoResultFound("A template type with resource type NETWORK "
                                f"not found in template {network_template_i.name}"
                                f" ({network_template_i.id})")

    except NoResultFound:
        log.debug("No network type to set.")

    for node_i in net_i.nodes:
        templates = get_types_by_attr(node_i, 'NODE', template_id, **kwargs)
        if len(templates) > 0:
            assign_type_to_resource(templates[0].id, 'NODE', node_i.id, **kwargs)
    for link_i in net_i.links:
        templates = get_types_by_attr(link_i, 'LINK', template_id, **kwargs)
        if len(templates) > 0:
            assign_type_to_resource(templates[0].id, 'LINK', link_i.id, **kwargs)

    for group_i in net_i.resourcegroups:
        templates = get_types_by_attr(group_i, 'GROUP', template_id, **kwargs)
        if len(templates) > 0:
            assign_type_to_resource(templates[0].id, 'GROUP', group_i.id, **kwargs)

    db.DBSession.flush()

@required_perms("edit_network")
def set_network_template(template_id, network_id, **kwargs):
    """
       Apply an existing template to a network.
       Used when a template has changed, and additional attributes
       must be added to the network's elements.
    """

    resource_types = []

    try:
        network_type = db.DBSession.query(ResourceType).filter(
            ResourceType.ref_key == 'NETWORK',
            ResourceType.network_id == network_id,
            ResourceType.type_id == TemplateType.id,
            TemplateType.template_id == template_id).one()
        resource_types.append(network_type)
    except NoResultFound:
        log.debug("No network type to set.")

    node_types = db.DBSession.query(ResourceType).filter(
        ResourceType.ref_key == 'NODE',
        ResourceType.node_id == Node.id,
        Node.network_id == network_id,
        ResourceType.type_id == TemplateType.id,
        TemplateType.template_id == template_id).all()

    link_types = db.DBSession.query(ResourceType).filter(
        ResourceType.ref_key == 'LINK',
        ResourceType.link_id == Link.id,
        Link.network_id == network_id,
        ResourceType.type_id == TemplateType.id,
        TemplateType.template_id == template_id).all()

    group_types = db.DBSession.query(ResourceType).filter(
        ResourceType.ref_key == 'GROUP',
        ResourceType.group_id == ResourceGroup.id,
        ResourceGroup.network_id == network_id,
        ResourceType.type_id == TemplateType.id,
        TemplateType.template_id == template_id).all()

    resource_types.extend(node_types)
    resource_types.extend(link_types)
    resource_types.extend(group_types)

    assign_types_to_resources(resource_types)

    log.debug("Finished setting network template")

@required_perms("edit_network")
def remove_template_from_network(network_id, template_id, remove_attrs, **kwargs):
    """
        Remove all resource types in a network relating to the specified
        template.
        remove_attrs ('Y' or 'N')
            Flag to indicate whether the attributes associated with the template
            types should be removed from the resources in the network. These will
            only be removed if they are not shared with another template on the network
    """

    try:
        network = db.DBSession.query(Network).filter(Network.id == network_id).one()
    except NoResultFound:
        raise HydraError("Network %s not found"%network_id)

    try:
        template = db.DBSession.query(Template).filter(Template.id == template_id).one()
    except NoResultFound:
        raise HydraError("Template %s not found"%template_id)

    type_ids = [tmpltype.id for tmpltype in template.get_types()]

    node_ids = [n.id for n in network.nodes]
    link_ids = [l.id for l in network.links]
    group_ids = [g.id for g in network.resourcegroups]

    if remove_attrs == 'Y':
        #find the attributes to remove
        resource_attrs_to_remove = _get_resources_to_remove(network, template)
        for n in network.nodes:
            resource_attrs_to_remove.extend(_get_resources_to_remove(n, template))
        for l in network.links:
            resource_attrs_to_remove.extend(_get_resources_to_remove(l, template))
        for g in network.resourcegroups:
            resource_attrs_to_remove.extend(_get_resources_to_remove(g, template))

        for ra in resource_attrs_to_remove:
            db.DBSession.delete(ra)

    resource_types = db.DBSession.query(ResourceType).filter(
        and_(or_(
            ResourceType.network_id==network_id,
            ResourceType.node_id.in_(node_ids),
            ResourceType.link_id.in_(link_ids),
            ResourceType.group_id.in_(group_ids),
        ), ResourceType.type_id.in_(type_ids))).all()

    for resource_type in resource_types:
        db.DBSession.delete(resource_type)

    db.DBSession.flush()

def _get_resources_to_remove(resource, template):
    """
        Given a resource and a template being removed, identify the resource attribtes
        which can be removed.
    """
    type_ids = [tmpltype.id for tmpltype in template.get_types()]

    node_attr_ids = dict([(ra.attr_id, ra) for ra in resource.attributes])
    attrs_to_remove = []
    attrs_to_keep = []
    for nt in resource.types:
        templatetype = nt.get_templatetype()
        if templatetype in type_ids:
            for ta in templatetype.typeattrs:
                if node_attr_ids.get(ta.attr_id):
                    attrs_to_remove.append(node_attr_ids[ta.attr_id])
        else:
            for ta in templatetype.typeattrs:
                if node_attr_ids.get(ta.attr_id):
                    attrs_to_keep.append(node_attr_ids[ta.attr_id])
    #remove any of the attributes marked for deletion as they are
    #marked for keeping based on being in another type.
    final_attrs_to_remove = set(attrs_to_remove) - set(attrs_to_keep)

    return list(final_attrs_to_remove)

def get_matching_resource_types(resource_type, resource_id, **kwargs):
    """
        Get the possible types of a resource by checking its attributes
        against all available types.

        @returns A list of TypeSummary objects.
    """
    resource_i = None
    if resource_type == 'NETWORK':
        resource_i = db.DBSession.query(Network).filter(Network.id == resource_id).one()
    elif resource_type == 'NODE':
        resource_i = db.DBSession.query(Node).filter(Node.id == resource_id).one()
    elif resource_type == 'LINK':
        resource_i = db.DBSession.query(Link).filter(Link.id == resource_id).one()
    elif resource_type == 'GROUP':
        resource_i = db.DBSession.query(ResourceGroup).filter(
            ResourceGroup.id == resource_id).one()

    matching_types = get_types_by_attr(resource_i, resource_type, **kwargs)
    return matching_types

@required_perms("edit_network")
def assign_types_to_resources(resource_types, template_id=None, **kwargs):
    """
        Assign new types to list of resources.
        This function checks if the necessary
        attributes are present and adds them if needed. Non existing attributes
        are also added when the type is already assigned. This means that this
        function can also be used to update resources, when a resource type has
        changed.
    """

    log.info("Setting types of %s resources", len(resource_types))
    #Remove duplicate values from types by turning it into a set
    type_ids = list(set([rt.type_id for rt in resource_types]))

    #Get the template ID from the incoming data. We need this so we can get the
    #correct types from the template hierarchy.
    #We assume here that this function can only be called in the context of
    #one template -- you can't send resource types from 2 templates.
    if template_id is None and resource_types[0].template_id is None:
        log.info("No template ID specified. Getting from type")
        db_type = _get_type(resource_types[0].type_id)
        template_id = db_type.template_id
        log.info("Template ID set to: %s", template_id)

    template_i = db.DBSession.query(Template).filter(Template.id == template_id).one()

    template_types = template_i.get_types()

    db_types = []
    for tt in template_types:
        if tt.id in type_ids:
            db_types.append(tt)

    type_lookup = {}
    for db_type in db_types:
        if type_lookup.get(db_type.id) is None:
            type_lookup[db_type.id] = db_type
    log.debug("Retrieved all the appropriate template types")
    res_types = []
    res_attrs = []
    res_scenarios = []

    net_id = None
    node_ids = []
    link_ids = []
    grp_ids = []
    for resource_type in resource_types:
        ref_id = resource_type.ref_id
        ref_key = resource_type.ref_key
        if resource_type.ref_key == 'NETWORK':
            net_id = ref_id
        elif resource_type.ref_key == 'NODE':
            node_ids.append(ref_id)
        elif resource_type.ref_key == 'LINK':
            link_ids.append(ref_id)
        elif resource_type.ref_key == 'GROUP':
            grp_ids.append(ref_id)
    if net_id:
        net = db.DBSession.query(Network).filter(Network.id == net_id).one()
    nodes = _get_nodes(node_ids)
    links = _get_links(link_ids)
    groups = _get_groups(grp_ids)
    for resource_type in resource_types:
        ref_id = resource_type.ref_id
        ref_key = resource_type.ref_key
        type_id = resource_type.type_id
        if ref_key == 'NETWORK':
            resource = net
        elif ref_key == 'NODE':
            resource = nodes[ref_id]
        elif ref_key == 'LINK':
            resource = links[ref_id]
        elif ref_key == 'GROUP':
            resource = groups[ref_id]

        ra, rt, rs = set_resource_type(resource, type_id, type_lookup, **kwargs)
        if rt is not None:
            res_types.append(rt)
        if len(ra) > 0:
            res_attrs.extend(ra)
        if len(rs) > 0:
            res_scenarios.extend(rs)

    log.debug("Retrieved all the appropriate resources")
    if len(res_types) > 0:
        new_types = db.DBSession.execute(ResourceType.__table__.insert(), res_types)

    new_ras = []
    if len(res_attrs) > 0:
        new_res_attrs = db.DBSession.execute(ResourceAttr.__table__.insert(), res_attrs)
        last_row_id = new_res_attrs.lastrowid or 0
        new_ras = db.DBSession.query(ResourceAttr).filter(
            and_(ResourceAttr.id >= last_row_id,
                 ResourceAttr.id < (last_row_id+len(res_attrs)))).all()

    ra_map = {}
    for new_ra in new_ras:
        ra_map[(new_ra.ref_key,
                new_ra.attr_id,
                new_ra.node_id,
                new_ra.link_id,
                new_ra.group_id,
                new_ra.network_id)] = new_ra.id

    for rs in res_scenarios:
        rs['resource_attr_id'] = ra_map[(rs['ref_key'],
                                         rs['attr_id'],
                                         rs['node_id'],
                                         rs['link_id'],
                                         rs['group_id'],
                                         rs['network_id'])]

    if len(res_scenarios) > 0:
        db.DBSession.execute(ResourceScenario.__table__.insert(), res_scenarios)

    #Make DBsession 'dirty' to pick up the inserts by doing a fake delete.
    db.DBSession.query(ResourceAttr).filter(ResourceAttr.attr_id is None).delete()

    ret_val = list(type_lookup.values())

    return ret_val

@required_perms('get_template')
def check_type_compatibility(type_1_id, type_2_id, **kwargs):
    """
        When applying a type to a resource, it may be the case that the resource already
        has an attribute specified in the new type, but the template which defines this
        pre-existing attribute has a different unit specification to the new template.

        This function checks for any situations where different types specify the same
        attributes, but with different units.
        args:
            type_1_id: The ID of the type to compare
            type_2_id: The ID of the type to compare
        returns:
            list of strings, with errors describing any incompatibilities
    """
    errors = []

    type_1 = _get_type(type_1_id)
    type_2 = _get_type(type_2_id)

    template_1_name = type_1.template.name
    template_2_name = type_2.template.name

    type_1_attrs = set([t.attr_id for t in type_1.typeattrs])
    type_2_attrs = set([t.attr_id for t in type_2.typeattrs])

    shared_attrs = type_1_attrs.intersection(type_2_attrs)

    if len(shared_attrs) == 0:
        return []

    type_1_dict = {}
    for t in type_1.typeattrs:
        if t.attr_id in shared_attrs:
            type_1_dict[t.attr_id] = t

    for ta in type_2.typeattrs:
        type_2_unit_id = ta.unit_id
        type_1_unit_id = type_1_dict[ta.attr_id].unit_id

        fmt_dict = {
            'template_1_name':template_1_name,
            'template_2_name':template_2_name,
            'attr_name':ta.attr.name,
            'type_1_unit_id':type_1_unit_id,
            'type_2_unit_id':type_2_unit_id,
            'type_name' :type_1.name
        }

        if type_1_unit_id is None and type_2_unit_id is not None:
            errors.append("Type %(type_name)s in template %(template_1_name)s"
                          " stores %(attr_name)s with no units, while template"
                          "%(template_2_name)s stores it with unit %(type_2_unit_id)s"%fmt_dict)
        elif type_1_unit_id is not None and type_2_unit_id is None:
            errors.append("Type %(type_name)s in template %(template_1_name)s"
                          " stores %(attr_name)s in %(type_1_unit_id)s."
                          " Template %(template_2_name)s stores it with no unit."%fmt_dict)
        elif type_1_unit_id != type_2_unit_id:
            errors.append("Type %(type_name)s in template %(template_1_name)s"
                          " stores %(attr_name)s in %(type_1_unit_id)s, while"
                          " template %(template_2_name)s stores it in %(type_2_unit_id)s"%fmt_dict)
        return errors

def _get_links(link_ids):
    links = []

    if len(link_ids) == 0:
        return {}

    block_size = 500
    limit = len(link_ids)
    lower = 0
    while lower < limit:
        upper = lower+block_size
        rs = db.DBSession.query(Link)\
                   .options(joinedload(Link.attributes))\
                   .options(joinedload(Link.types))\
                   .filter(Link.id.in_(link_ids[lower:upper])).all()
        log.debug("Retrieved %s links", len(rs))
        links.extend(rs)
        lower = upper


    link_dict = {}

    for l in links:
        l.ref_id = l.id
        l.ref_key = 'LINK'
        link_dict[l.id] = l

    return link_dict

def _get_nodes(node_ids):
    nodes = []

    if len(node_ids) == 0:
        return {}

    block_size = 500
    limit = len(node_ids)
    lower = 0
    while lower < limit:
        upper = lower+block_size
        rs = db.DBSession.query(Node)\
                   .options(joinedload(Node.attributes))\
                   .options(joinedload(Node.types))\
                   .filter(Node.id.in_(node_ids[lower:upper])).all()
        log.debug("Retrieved %s nodes", len(rs))
        nodes.extend(rs)
        lower = upper

    node_dict = {}

    for n in nodes:
        n.ref_id = n.id
        n.ref_key = 'NODE'
        node_dict[n.id] = n

    return node_dict

def _get_groups(group_ids):
    groups = []

    if len(group_ids) == 0:
        return {}

    block_size = 500
    limit = len(group_ids)
    lower = 0
    while lower < limit:
        upper = lower+block_size
        rs = db.DBSession.query(ResourceGroup)\
                   .options(joinedload(ResourceGroup.attributes))\
                   .options(joinedload(ResourceGroup.types))\
                   .filter(ResourceGroup.id.in_(group_ids[lower:upper])).all()
        log.debug("Retrieved %s groups", len(rs))
        groups.extend(rs)
        lower = upper

    group_dict = {}

    for g in groups:
        g.ref_id = g.id
        g.ref_key = 'GROUP'
        group_dict[g.id] = g

    return group_dict

@required_perms("edit_network")
def assign_type_to_resource(type_id, resource_type, resource_id, **kwargs):
    """Assign new type to a resource. This function checks if the necessary
    attributes are present and adds them if needed. Non existing attributes
    are also added when the type is already assigned. This means that this
    function can also be used to update resources, when a resource type has
    changed.
    """

    if resource_type == 'NETWORK':
        resource = db.DBSession.query(Network).filter(Network.id == resource_id).one()
    elif resource_type == 'NODE':
        resource = db.DBSession.query(Node).filter(Node.id == resource_id).one()
    elif resource_type == 'LINK':
        resource = db.DBSession.query(Link).filter(Link.id == resource_id).one()
    elif resource_type == 'GROUP':
        resource = db.DBSession.query(ResourceGroup).filter(ResourceGroup.id == resource_id).one()

    res_attrs, res_type, res_scenarios = set_resource_type(resource, type_id, **kwargs)

    type_i = _get_type(type_id)

    if resource_type != type_i.resource_type:
        raise HydraError("Cannot assign a %s type to a %s"%
                         (type_i.resource_type, resource_type))

    if res_type is not None:
        db.DBSession.bulk_insert_mappings(ResourceType, [res_type])

    if len(res_attrs) > 0:
        db.DBSession.bulk_insert_mappings(ResourceAttr, res_attrs)

    if len(res_scenarios) > 0:
        db.DBSession.bulk_insert_mappings(ResourceScenario, res_scenarios)

    #Make DBsession 'dirty' to pick up the inserts by doing a fake delete.
    db.DBSession.query(Attr).filter(Attr.id is None).delete()

    db.DBSession.flush()

    return_type = _get_type(type_id)

    return return_type

@required_perms("edit_network")
def set_resource_type(resource, type_id, types={}, **kwargs):
    """
        Set this resource to be a certain type.
        Type objects (a dictionary keyed on type_id) may be
        passed in to save on loading.
        This function does not call save. It must be done afterwards.
        New resource attributes are added to the resource if the template
        requires them. Resource attributes on the resource but not used by
        the template are not removed.
        @returns list of new resource attributes
        ,new resource type object
    """

    #get the resource#s network ID:
    if kwargs.get('network_id') is not None:
        network_id = kwargs['network_id']
    elif isinstance(resource, Network):
        network_id = resource.id
    elif resource.network_id:
        network_id = resource.network_id
    elif resource.network:
        network_id = resource.network.id

    child_template_id = kwargs.get('child_template_id')
    if kwargs.get('child_template_id') is None:
        if network_id is not None:
            child_template_id = get_network_template(network_id, type_id)

    ref_key = resource.ref_key

    existing_attr_ids = []
    for res_attr in resource.attributes:
        existing_attr_ids.append(res_attr.attr_id)

    if type_id in types:
        type_i = types[type_id]
    else:
        type_i = _get_type(type_id)

    type_attrs = dict()
    for typeattr in type_i.typeattrs:
        type_attrs[typeattr.attr_id] = {
            'is_var':typeattr.attr_is_var,
            'default_dataset_id': typeattr.default_dataset.id if typeattr.default_dataset else None
        }

    # check if attributes exist
    missing_attr_ids = set(type_attrs.keys()) - set(existing_attr_ids)

    # add attributes if necessary
    new_res_attrs = []

    #This is a dict as the length of the list may not match the new_res_attrs
    #Keyed on attr_id, as resource_attr_id doesn't exist yet, and there should only
    #be one attr_id per template.
    new_res_scenarios = {}
    for attr_id in missing_attr_ids:
        ra_dict = dict(
            ref_key=ref_key,
            attr_id=attr_id,
            attr_is_var=type_attrs[attr_id]['is_var'],
            node_id=resource.id if ref_key == 'NODE' else None,
            link_id=resource.id if ref_key == 'LINK' else None,
            group_id=resource.id if ref_key == 'GROUP' else None,
            network_id=resource.id if ref_key == 'NETWORK' else None,
        )
        new_res_attrs.append(ra_dict)



        if type_attrs[attr_id]['default_dataset_id'] is not None:
            if hasattr(resource, 'network'):
                for s in resource.network.scenarios:

                    if new_res_scenarios.get(attr_id) is None:
                        new_res_scenarios[attr_id] = {}

                    new_res_scenarios[attr_id][s.id] = dict(
                        dataset_id=type_attrs[attr_id]['default_dataset_id'],
                        scenario_id=s.id,
                        #Not stored in the DB, but needed to connect the RA ID later.
                        attr_id=attr_id,
                        ref_key=ref_key,
                        node_id=ra_dict['node_id'],
                        link_id=ra_dict['link_id'],
                        group_id=ra_dict['group_id'],
                        network_id=ra_dict['network_id'],
                    )


    resource_type = None
    for rt in resource.types:

        if rt.type_id == type_i.id:
            break

        errors = check_type_compatibility(rt.type_id, type_i.id, **kwargs)
        if len(errors) > 0:
            raise HydraError("Cannot apply type %s to resource %s as it "
                             "conflicts with type %s. Errors are: %s"
                             %(type_i.name, resource.get_name(),
                               rt.get_templatetype().name, ','.join(errors)))
    else:
        # add type to tResourceType if it doesn't exist already
        resource_type = dict(
            node_id=resource.id if ref_key == 'NODE' else None,
            link_id=resource.id if ref_key == 'LINK' else None,
            group_id=resource.id if ref_key == 'GROUP' else None,
            network_id=resource.id if ref_key == 'NETWORK' else None,
            ref_key=ref_key,
            type_id=type_id,
            child_template_id=child_template_id
        )

    return new_res_attrs, resource_type, new_res_scenarios

@required_perms("edit_network")
def remove_type_from_resource( type_id, resource_type, resource_id, **kwargs):
    """
        Remove a resource type trom a resource
    """
    node_id = resource_id if resource_type == 'NODE' else None
    link_id = resource_id if resource_type == 'LINK' else None
    group_id = resource_id if resource_type == 'GROUP' else None

    resourcetype = db.DBSession.query(ResourceType).filter(
        ResourceType.type_id == type_id,
        ResourceType.ref_key == resource_type,
        ResourceType.node_id == node_id,
        ResourceType.link_id == link_id,
        ResourceType.group_id == group_id).one()

    db.DBSession.delete(resourcetype)
    db.DBSession.flush()

    return 'OK'


@required_perms('get_network')
def validate_attr(resource_attr_id, scenario_id, template_id=None, **kwargs):
    """
        Check that a resource attribute satisfies the requirements of all the types of the
        resource.
    """
    rs = db.DBSession.query(ResourceScenario).\
                        filter(ResourceScenario.resource_attr_id == resource_attr_id,
                               ResourceScenario.scenario_id == scenario_id).options(
                                   joinedload(ResourceScenario.resourceattr)).options(
                                       joinedload(ResourceScenario.dataset)
                                   ).one()

    error = None

    try:
        validate_resourcescenario(rs, template_id)
    except HydraError as e:

        error = JSONObject(dict(
            ref_key=rs.resourceattr.ref_key,
            ref_id=rs.resourceattr.get_resource_id(),
            ref_name=rs.resourceattr.get_resource().get_name(),
            resource_attr_id=rs.resource_attr_id,
            attr_id=rs.resourceattr.attr.id,
            attr_name=rs.resourceattr.attr.name,
            dataset_id=rs.dataset_id,
            scenario_id=scenario_id,
            template_id=template_id,
            error_text=e.args[0]))
    return error

@required_perms('get_network')
def validate_attrs(resource_attr_ids, scenario_id, template_id=None, **kwargs):
    """
        Check that multiple resource attribute satisfy the requirements of the types of resources to
        which the they are attached.
    """
    multi_rs = db.DBSession.query(ResourceScenario).\
                            filter(ResourceScenario.resource_attr_id.in_(resource_attr_ids),\
                                   ResourceScenario.scenario_id == scenario_id).\
                                   options(joinedload(ResourceScenario.resourceattr)).\
                                   options(joinedload(ResourceScenario.dataset)).all()

    errors = []
    for rs in multi_rs:
        try:
            validate_resourcescenario(rs, template_id)
        except HydraError as e:

            error = dict(
                ref_key=rs.resourceattr.ref_key,
                ref_id=rs.resourceattr.get_resource_id(),
                ref_name=rs.resourceattr.get_resource().get_name(),
                resource_attr_id=rs.resource_attr_id,
                attr_id=rs.resourceattr.attr.id,
                attr_name=rs.resourceattr.attr.name,
                dataset_id=rs.dataset_id,
                scenario_id=scenario_id,
                template_id=template_id,
                error_text=e.args[0])

            errors.append(error)

    return errors

@required_perms('get_network')
def validate_scenario(scenario_id, template_id=None, **kwargs):
    """
        Check that the requirements of the types of resources in a scenario are
        correct, based on the templates in a network. If a template is specified,
        only that template will be checked.
    """
    scenario_rs = db.DBSession.query(ResourceScenario).filter(
                ResourceScenario.scenario_id==scenario_id)\
                .options(joinedload(ResourceScenario.resourceattr))\
                .options(joinedload(ResourceScenario.dataset)).all()

    errors = []
    for rs in scenario_rs:
        try:
            validate_resourcescenario(rs, template_id)
        except HydraError as e:

            error = dict(
                ref_key=rs.resourceattr.ref_key,
                ref_id=rs.resourceattr.get_resource_id(),
                ref_name=rs.resourceattr.get_resource().get_name(),
                resource_attr_id=rs.resource_attr_id,
                attr_id=rs.resourceattr.attr.id,
                attr_name=rs.resourceattr.attr.name,
                dataset_id=rs.dataset_id,
                scenario_id=scenario_id,
                template_id=template_id,
                error_text=e.args[0])
            errors.append(error)

    return errors


def validate_resourcescenario(resourcescenario, template_id=None, **kwargs):
    """
        Perform a check to ensure a resource scenario's datasets are correct given what the
        definition of that resource (its type) specifies.
    """
    res = resourcescenario.resourceattr.get_resource()

    types = res.types

    dataset = resourcescenario.dataset

    if len(types) == 0:
        return

    if template_id is not None:
        if template_id not in [r.get_templatetype().template_id for r in res.types]:
            raise HydraError("Template %s is not used for resource attribute %s in scenario %s"%\
                             (template_id, resourcescenario.resourceattr.attr.name,
                              resourcescenario.scenario.name))

    #Validate against all the types for the resource
    for resourcetype in types:
        #If a specific type has been specified, then only validate
        #against that type and ignore all the others
        if template_id is not None:
            if resourcetype.get_templatetype().template_id != template_id:
                continue
        #Identify the template types for the template
        tmpltype = resourcetype.get_templatetype()
        for ta in tmpltype.typeattrs:
            #If we find a template type which mactches the current attribute.
            #we can do some validation.
            if ta.attr_id == resourcescenario.resourceattr.attr_id:
                if ta.data_restriction:
                    log.debug("Validating against %s", ta.data_restriction)
                    validation_dict = json.loads(ta.data_restriction)
                    dataset_util.validate_value(validation_dict, dataset.get_val())

@required_perms('get_network')
def validate_network(network_id, template_id, scenario_id=None, **kwargs):
    """
        Given a network, scenario and template, ensure that all the nodes, links & groups
        in the network have the correct resource attributes as defined by
        the types in the template.
        Also ensure valid entries in tresourcetype.
        This validation will not fail if a resource has more than the required type,
        but will fail if it has fewer or if any attribute has a
        conflicting dimension or unit.
    """

    network = db.DBSession.query(Network).filter(
        Network.id == network_id).options(noload(Network.scenarios)).first()

    if network is None:
        raise HydraError("Could not find network %s"%(network_id))

    resource_scenario_dict = {}
    if scenario_id is not None:
        scenario = db.DBSession.query(Scenario).filter(Scenario.id == scenario_id).first()

        if scenario is None:
            raise HydraError("Could not find scenario %s"%(scenario_id,))

        for rs in scenario.resourcescenarios:
            resource_scenario_dict[rs.resource_attr_id] = rs

    template = db.DBSession.query(Template).filter(
        Template.id == template_id).first()

    if template is None:
        raise HydraError("Could not find template %s"%(template_id,))

    resource_type_defs = {
        'NETWORK' : {},
        'NODE'    : {},
        'LINK'    : {},
        'GROUP'   : {},
    }
    for tt in template.get_types():
        resource_type_defs[tt.resource_type][tt.id] = tt

    errors = []
    #Only check if there are type definitions for a network in the template.
    if resource_type_defs.get('NETWORK'):
        net_types = resource_type_defs['NETWORK']
        errors.extend(validate_resource(network, net_types, resource_scenario_dict))

    #check all nodes
    if resource_type_defs.get('NODE'):
        node_types = resource_type_defs['NODE']
        for node in network.nodes:
            errors.extend(validate_resource(node, node_types, resource_scenario_dict))

    #check all links
    if resource_type_defs.get('LINK'):
        link_types = resource_type_defs['LINK']
        for link in network.links:
            errors.extend(validate_resource(link, link_types, resource_scenario_dict))

    #check all groups
    if resource_type_defs.get('GROUP'):
        group_types = resource_type_defs['GROUP']
        for group in network.resourcegroups:
            errors.extend(validate_resource(group, group_types, resource_scenario_dict))

    return errors

def validate_resource(resource, tmpl_types, resource_scenarios=[], **kwargs):
    errors = []
    resource_type = None

    #No validation required if the link has no type.
    if len(resource.types) == 0:
        return []

    for rt in resource.types:
        if tmpl_types.get(rt.type_id) is not None:
            resource_type = tmpl_types[rt.type_id]
            break
    else:
        errors.append("Type %s not found on %s %s"%
                      (tmpl_types, resource_type, resource.get_name()))

    ta_dict = {}
    for ta in resource_type.typeattrs:
        ta_dict[ta.attr_id] = ta

    #Make sure the resource has all the attributes specified in the tempalte
    #by checking whether the template attributes are a subset of the resource
    #attributes.
    type_attrs = set([ta.attr_id for ta in resource_type.typeattrs])

    resource_attrs = set([ra.attr_id for ra in resource.attributes])

    if not type_attrs.issubset(resource_attrs):
        for ta in type_attrs.difference(resource_attrs):
            attr = db.DBSession.query(Attr).filter(Attr.id==ta).one()
            errors.append("Resource %s does not have attribute %s"%
                          (resource.get_name(), attr.name))

    resource_attr_ids = set([ra.id for ra in resource.attributes])
    #if data is included, check to make sure each dataset conforms
    #to the boundaries specified in the template: i.e. that it has
    #the correct dimension and (if specified) unit.
    if len(resource_scenarios) > 0:
        for ra_id in resource_attr_ids:
            rs = resource_scenarios.get(ra_id)
            if rs is None:
                continue
            attr_name = rs.resourceattr.attr.name
            rs_unit_id = rs.dataset.unit_id

            rs_dimension_id = units.get_dimension_by_unit_id(rs_unit_id,
                                                             do_accept_unit_id_none=True).id


            attr = db.DBSession.query(Attr).filter(Attr.id==rs.resourceattr.attr_id).one()
            type_dimension_id = attr.dimension_id
            type_unit_id = ta_dict[rs.resourceattr.attr_id].unit_id

            if rs_dimension_id != type_dimension_id:
                errors.append("Dimension mismatch on %s %s, attribute %s: "
                              "%s on attribute, %s on type"%
                              (resource.ref_key, resource.get_name(), attr_name,
                               rs_dimension_id, type_dimension_id))

            if type_unit_id is not None:
                if rs_unit_id != type_unit_id:
                    errors.append("Unit mismatch on attribute %s. "
                                  "%s on attribute, %s on type"%
                                  (attr_name, rs_unit_id, type_unit_id))
    if len(errors) > 0:
        log.warning(errors)

    return errors



"""
    A simple cache system for storing such things as project hierarchies and templates.
    By default uses diskcache for simpler setup and backward compatibility
    unless 'memcached' is set in the 'cache' section of the
    config, in which case use that.
"""
import logging
import datetime
from hydra_base import config as hydraconfig
import tempfile

log = logging.getLogger(__name__)
global cache

def _init_diskcache():
    log.info("Using diskcache for caching.")
    global cache
    import diskcache as dc
    cache = dc.Cache(tempfile.gettempdir())

if hydraconfig.get('cache', 'type') != "memcached":
    _init_diskcache()

elif hydraconfig.get('cache', 'type') == 'memcached':
    try:
        import pylibmc
        host = hydraconfig.get('cache', 'host', '127.0.0.1')
        port = hydraconfig.get('cache', 'port', 31211)
        cache = pylibmc.Client([f"{host}:{port}"], binary=True)

        # Check if Memcached server is reachable by setting a test key
        test_key = "__connection_test__"
        #pick a unique key based on the time
        test_value = datetime.datetime.toordinal(datetime.datetime.now())
        try:
            cache.set(test_key, test_value, 1)
            cache.get(test_key)
            log.info("Connected to memcached server.")
        except Exception:
            raise ConnectionError("Memcached server not responding.")

    except (ModuleNotFoundError, ConnectionError) as e:
        if isinstance(e, ModuleNotFoundError):
            log.warning("Unable to find pylibmc. Defaulting to diskcache.")
        else:
            log.warning("Memcached server not reachable. Defaulting to diskcache.")

        _init_diskcache()

def clear_cache():
    if hasattr(cache, 'flush_all'):
        cache.flush_all() # memcache
    else:
        cache.clear() # diskcache



from abc import (
    ABC,
    abstractmethod
)

class DatasetManager(ABC):
    """
    Base class for DatasetManagers.
    This acts as a descriptor in the model.Dataset class and
    subclasses must implement xxx_storage_location() methods
    in addition to the __get__/__set__ descriptor protocol
    """
    @abstractmethod
    def __get__(self, dataset, dtype):
        pass

    @abstractmethod
    def __set__(self, dataset, value):
        pass

    @abstractmethod
    def get_storage_location(self, dataset):
        pass

    @abstractmethod
    def set_storage_location(self, dataset, location):
        pass

    @abstractmethod
    def delete_storage_location(self, dataset):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

__all__ = ['Note']

class Note(Base, Inspect, PermissionControlled):
    """
        A note is an arbitrary piece of text which can be applied
        to any resource. A note is NOT scenario dependent. It is applied
        directly to resources. A note can be applied to a scenario.
    """

    __tablename__='tNote'

    id = Column(Integer(), primary_key=True, nullable=False)
    ref_key = Column(String(60),  nullable=False, index=True)
    value = Column(LargeBinary(),  nullable=True)
    created_by = Column(Integer(), ForeignKey('tUser.id'))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    scenario_id = Column(Integer(), ForeignKey('tScenario.id'),  index=True, nullable=True)
    project_id = Column(Integer(), ForeignKey('tProject.id'),  index=True, nullable=True)
    network_id  = Column(Integer(),  ForeignKey('tNetwork.id'), index=True, nullable=True,)
    node_id     = Column(Integer(),  ForeignKey('tNode.id'), index=True, nullable=True)
    link_id     = Column(Integer(),  ForeignKey('tLink.id'), index=True, nullable=True)
    group_id    = Column(Integer(),  ForeignKey('tResourceGroup.id'), index=True, nullable=True)

    scenario = relationship('Scenario', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')
    node = relationship('Node', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')
    link = relationship('Link', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')
    group = relationship('ResourceGroup', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')
    network = relationship('Network', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')
    project = relationship('Project', backref=backref('notes', uselist=True, cascade="all, delete-orphan"), uselist=True, lazy='joined')

    _parents  = ['tScenario', 'tNode', 'tLink', 'tProject', 'tNetwork', 'tResourceGroup']
    _children = []

    def set_ref(self, ref_key, ref_id):
        """
            Using a ref key and ref id set the
            reference to the appropriate resource type.
        """
        if ref_key == 'NETWORK':
            self.network_id = ref_id
        elif ref_key == 'NODE':
            self.node_id = ref_id
        elif ref_key == 'LINK':
            self.link_id = ref_id
        elif ref_key == 'GROUP':
            self.group_id = ref_id
        elif ref_key == 'SCENARIO':
            self.scenario_id = ref_id
        elif ref_key == 'PROJECT':
            self.project_id = ref_id

        else:
            raise HydraError("Ref Key %s not recognised."%ref_key)

    def get_ref_id(self):

        """
            Return the ID of the resource to which this not is attached
        """
        if self.ref_key == 'NETWORK':
            return self.network_id
        elif self.ref_key == 'NODE':
            return self.node_id
        elif self.ref_key == 'LINK':
            return self.link_id
        elif self.ref_key == 'GROUP':
            return self.group_id
        elif self.ref_key == 'SCENARIO':
            return self.scenario_id
        elif self.ref_key == 'PROJECT':
            return self.project_id

    def get_ref(self):
        """
            Return the ID of the resource to which this not is attached
        """
        if self.ref_key == 'NETWORK':
            return self.network
        elif self.ref_key == 'NODE':
            return self.node
        elif self.ref_key == 'LINK':
            return self.link
        elif self.ref_key == 'GROUP':
            return self.group
        elif self.ref_key == 'SCENARIO':
            return self.scenario
        elif self.ref_key == 'PROJECT':
            return self.project


